import numpy as np
import sys
import os
from collections import OrderedDict
from scipy.sparse import csr_matrix, dia_matrix, coo_matrix
from scipy.sparse import vstack as sparse_vstack
from scipy.sparse import hstack as sparse_hstack
from scipy.sparse import triu, tril
import tables
from intervaltree import IntervalTree, Interval

import gzip

# try to import pandas if exists
try:
    import pandas as pd
    pandas = True
except ImportError:
    pandas = False


class hiCMatrix:
    """
    Class to handle HiC matrices
    contains routines to get intrachromosomal distances
    get sub matrices by chrname.
    """

    def __init__(self, matrixFile=None, file_format=None, skiprows=None, chrnameList=None, bplimit=None):
        self.correction_factors = None  # this value is set in case a matrix was iteratively corrected
        self.non_homogeneous_warning_already_printed = False
        self.distance_counts = None  # only defined when getCountsByDistance is called
        self.bin_size = None
        self.bin_size_homogeneous = None  # track if the bins are equally spaced or not

        if matrixFile:
            self.nan_bins = np.array([])
            if not file_format:
                if matrixFile[-4:] == ".npz":
                    file_format = 'npz'
                elif matrixFile[-3:] == ".h5":
                    file_format = 'h5'
                elif matrixFile[-3:] == '.gz':
                    file_format = 'dekker'
                # by default assume that the matrix file_format is hd5
                else:
                    file_format = 'h5'

            if file_format == 'h5' or file_format == 'hicexplorer':
                self.matrix, self.cut_intervals, self.nan_bins, self.distance_counts, self.correction_factors = \
                    hiCMatrix.load_h5(matrixFile)
                self.restoreMaskedBins()

            elif file_format == 'npz':
                self.matrix, self.cut_intervals, self.nan_bins, self.distance_counts, self.correction_factors = \
                    hiCMatrix.load_npz(matrixFile)
                self.restoreMaskedBins()

            elif file_format == 'dekker':
                self.cut_intervals = self.getDekkerBins(matrixFile)
                if not skiprows:
                    # +1 to also skip the column labels
                    skiprows = len(self.header) + 1
                self.matrix = csr_matrix(
                    np.loadtxt(matrixFile,
                               skiprows=skiprows,
                               usecols=range(1, len(self.cut_intervals) + 1)))
                """
                # convert nans to zeros
                self.matrix.data[np.isnan(self.matrix.data)] = 0
                self.matrix.eliminate_zeros()
                # mask zero count bins:
                row_sum = np.asarray(self.matrix.sum(axis=1)).flatten()
                self.maskBins(np.flatnonzero(row_sum==0))
                """
            elif file_format == 'lieberman':  # lieberman format needs additional arguments : chrnameList
                lieberman_data = self.getLiebermanBins(filenameList=matrixFile, chrnameList=chrnameList)
                self.cut_intervals = lieberman_data['cut_intervals']
                self.matrix = lieberman_data['matrix']

            else:
                exit("matrix format not known.")

            self.interval_trees, self.chrBinBoundaries = \
                self.intervalListToIntervalTree(self.cut_intervals)

    @staticmethod
    def load_h5(matrix_filename):
        """
        Loads a matrix stored in h5 format
        :param matrix_filename:
        :return: matrix, cut_intervals, nan_bins, distance_counts, correction_factors
        """
        with tables.open_file(matrix_filename) as f:
            parts = {}
            for matrix_part in ('data', 'indices', 'indptr', 'shape'):
                parts[matrix_part] = getattr(f.root.matrix, matrix_part).read()

            matrix = csr_matrix(tuple([parts['data'], parts['indices'], parts['indptr']]),
                                shape=parts['shape'])
            matrix = hiCMatrix.fillLowerTriangle(matrix)
            # get intervals
            intvals = {}
            for interval_part in ('chr_list', 'start_list', 'end_list', 'extra_list'):
                intvals[interval_part] = getattr(f.root.intervals, interval_part).read()
            cut_intervals = zip(intvals['chr_list'], intvals['start_list'], intvals['end_list'], intvals['extra_list'])
            assert len(cut_intervals) == matrix.shape[0], \
                "Error loading matrix. Length of bin intervals ({}) is different than the " \
                "size of the matrix ({})".format(len(cut_intervals), matrix.shape[0])

            # get nan_bins
            if hasattr(f.root, 'nan_bins'):
                nan_bins = f.root.nan_bins.read()
            else:
                nan_bins = np.array([])

            # get correction factors
            if hasattr(f.root, 'correction_factors'):
                correction_factors = f.root.correction_factors.read()
                assert len(correction_factors) == matrix.shape[0], \
                    "Error loading matrix. Length of correction factors does not" \
                    "match size of matrix"
            else:
                correction_factors = None

            # get correction factors
            if hasattr(f.root, 'distance_counts'):
                distance_counts = f.root.correction_factors.read()
            else:
                distance_counts = None

            return matrix, cut_intervals, nan_bins, distance_counts, correction_factors

    @staticmethod
    def load_npz(matrixFile):
        _ma = np.load(matrixFile)
        matrix = hiCMatrix.fillLowerTriangle(_ma['matrix'].tolist())
        if 'dist_counts' not in _ma:
            distance_counts = None
        else:
            distance_counts = _ma['dist_counts'].tolist()

        cut_intervals = zip(_ma['chrNameList'], _ma['startList'],
                            _ma['endList'], _ma['extraList'])

        assert len(cut_intervals) == matrix.shape[0], \
            "Corrupted matrix file. Matrix size and " \
            "matrix bin definitions do not correspond"
        if 'nan_bins' in _ma.keys():
            nan_bins = _ma['nan_bins']
        else:
            nan_bins = np.array([])

        if 'correction_factors' in _ma.keys():
            try:
                # None value
                # for correction_factors is saved by numpy
                # as: array(None, dtype=object)
                # Thus, to get the original None value
                # the first item of the array is taken.
                _ma['correction_factors'][0]
                assert len(_ma['correction_factors']) == matrix.shape[0], \
                    "length of correction factors and length of matrix are different."
                correction_factors = _ma['correction_factors']
            except IndexError:
                correction_factors = None
        else:
            correction_factors = None

        return matrix, cut_intervals, nan_bins, distance_counts, correction_factors

    @staticmethod
    def fillLowerTriangle(matrix):
        """
        checks if the matrix is complete or if only half of the matrix was saved.
        Returns a whole matrix.

        >>> from scipy.sparse import csr_matrix
        >>> A = csr_matrix(np.array([[12,5,3,2,0],[0,11,4,1,1],
        ... [0,0,9,6,0], [0,0,0,10,0], [0,0,0,0,0]]))
        >>> B = hiCMatrix.fillLowerTriangle(A)
        >>> B.todense()
        matrix([[12,  5,  3,  2,  0],
                [ 5, 11,  4,  1,  1],
                [ 3,  4,  9,  6,  0],
                [ 2,  1,  6, 10,  0],
                [ 0,  1,  0,  0,  0]])

        """
        if tril(matrix, k=-1).sum() == 0:
            # this case means that the lower triangle of the
            # symmetric matrix (below the main diagonal)
            # is zero. In this case, replace the lower
            # triangle using the upper triangle
            matrix = matrix + triu(matrix, 1).T

        return matrix

    def setMatrix(self, matrix, cut_intervals):
        """
        Initialize a matrix with a given matrix
        and cut_intervals. Mostly useful for
        testing.
        """

        # check that the matrix is squared
        if matrix.shape[0] != matrix.shape[1]:
            raise Exception("Matrix is not squared. Shape is {}".format(matrix.shape))
        self.matrix = matrix
        self.cut_intervals = cut_intervals
        self.interval_trees, self.chrBinBoundaries = \
            self.intervalListToIntervalTree(self.cut_intervals)

    def getBinSize(self):
        """
        estimates the bin size. In case the bin size
        is not equal for all bins (maybe except for the
        bin at the en of the chromosomes) a warning is issued.
        In case of uneven bins, the median is returned.
        """
        if self.bin_size is None:
            chrom, start, end, extra = zip(*self.cut_intervals)
            median = int(np.median(np.diff(start)))
            diff = np.array(end) - np.array(start)
            # check if the bin size is homogeneous
            if len(np.flatnonzero(diff != median)) > (len(diff) * 0.01):
                self.bin_size_homogeneous = False
                if self.non_homogeneous_warning_already_printed is False:
                    sys.stderr.write('WARNING: bin size is not homogeneous. Median {}\n'.format(median))
                    self.non_homogeneous_warning_already_printed = True
            self.bin_size = median
        return self.bin_size

    def getDekkerBins(self, fileName):
        """
        Reads a gziped HiC matrix in Decker's format
        This format has row and column headers and comments
        with the prefix '#'. The following code skips the
        header while saving it in the self.header list
        and then reads and returns the column headers.
        The headers have the format bin_id|species|chrom:start-end
        for example 'bin1000000|dm3|chr4:1-50001'
        """

        i = 0
        self.header = []
        try:
            for line in gzip.open(fileName, 'r').readlines():
                if line.startswith("#"):
                    self.header.append(line)
                    continue
                i += 1
                if i == 1:  # this is the first line that is not
                            # a comment which contains the column headers
                            # that are equal to the row headers
                    colLabels = line.split("\t")
                    nameList = []
                    startList = []
                    endList = []
                    binIdList = []

                    for lab in colLabels[1:]:
                        (binId, species, position) = lab.split("|")
                        (chrName, pos) = position.split(":")
                        (start, end) = pos.split("-")
                        binIdList.append(binId)
                        nameList.append(chrName)
                        startList.append(int(start))
                        endList.append(int(end))
                    break

        except IOError:
            print "Error reading {}.\nDoes the file exists? "
            "Is it gzipped?".format(fileName)
            exit(1)

        return zip(nameList, startList, endList, binIdList)

    def getLiebermanBins(self, filenameList, chrnameList, pandas=pandas):
        """
        Reads a list of txt file in liberman's format and returns
        cut intervals and matrix. Each file is seperated by chr name
        and contains: locus1,locus2,and contact score seperated by tab.
        """

        # Create empty row, col, value for the matrix

        row = np.array([]).astype("int")
        col = np.array([]).astype("int")
        value = np.array([])
        cut_intervals = []
        dim = 0
        # for each chr, append the row, col, value to the first one. Extend the dim
        for i in range(0, len(filenameList)):
            if pandas is True:
                chrd = pd.read_csv(filenameList[i], sep="\t", header=None)
                chrdata = chrd.as_matrix()
            else:
                print "Pandas unavailable. Reading files using numpy (slower).."
                chrdata = np.loadtxt(filenameList[i])

            # define resolution as the median of the difference of the rows
            # in the data table.

            resolution = np.median(np.diff(np.unique(np.sort(chrdata[:, 1]))))

            chrcol = (chrdata[:, 1] / resolution).astype(int)
            chrrow = (chrdata[:, 0] / resolution).astype(int)

            chrdim = max(max(chrcol), max(chrrow)) + 1
            row = np.concatenate([row, chrrow + dim])
            col = np.concatenate([col, chrcol + dim])
            value = np.concatenate([value, chrdata[:, 2]])
            dim += chrdim

            for _bin in range(chrdim):
                cut_intervals.append((chrnameList[i], _bin * resolution, (_bin + 1) * resolution, 0))

        final_mat = coo_matrix((value, (row, col)), shape=(dim, dim))
        lieberman_data = dict(cut_intervals=cut_intervals, matrix=final_mat)
        return lieberman_data

    def getMatrix(self):
        matrix = self.matrix.todense()
        if len(self.nan_bins):
            # to set NaN values the matrix type has to be
            # float. Corrected matrices are of float
            # type while uncorrected matrices are of
            # of int type
            if np.issubdtype(self.matrix, 'float') is False:
                matrix = matrix.astype(float)
            matrix[self.nan_bins, :] = np.nan
            matrix[:, self.nan_bins] = np.nan

        return matrix

    def getChrBinRange(self, chrName):
        """
        Given a chromosome name,
        This functions return the start and end bin indices in the matrix
        """
        return self.chrBinBoundaries[chrName]

    def getChrNames(self):
        """
        returns the names of the chromosomes
        present in the matrix
        """
        return self.chrBinBoundaries.keys()

    def getBinPos(self, binIndex):
        """
        given a bin, it returns the chromosome name,
        start position and end position
        """
        return self.cut_intervals[binIndex]

    def getRegionBinRange(self, chrname, startpos, endpos):
        """
        Given a chromosome region, this function returns
        the bin indices that overlap with such region.
        """

        try:
            self.interval_trees[chrname]
        except KeyError:
            """
            print "chromosome: {} name not found in matrix".format(chrName)
            print "valid names are:"
            print interval_trees.keys()
            """
            return None
        try:
            startpos = int(startpos)
            endpos = int(endpos)
        except:
            print "{} or {}  are not valid " \
                "position values.".format(startpos, endpos)
            exit()

        try:
            startbin = sorted(self.interval_trees[chrname][startpos:startpos + 1])[0].data
            endbin = sorted(self.interval_trees[chrname][endpos:endpos + 1])[0].data
        except IndexError:
            return None

        return startbin, endbin

    @staticmethod
    def getDistList(rows, cols, cut_intervals):
        """
            Given a list of rows and cols
            an array is returned containing
            the genomic distance between
            each element of the row array
            with each element of the col array.
            -1 is returned for inter-chromosomal
            interactions.

            A matching list containing the chromosome name
            is also returned

        >>> from scipy.sparse import coo_matrix
        >>> row, col = np.triu_indices(5)
        >>> cut_intervals = [('a', 0, 10, 1), ('a', 10, 20, 1),
        ... ('a', 20, 30, 1), ('a', 30, 40, 1), ('b', 40, 50, 1)]
        >>> dist_list, chrom_list = hiCMatrix.getDistList(row, col,
        ... cut_intervals)
        >>> coo_matrix((dist_list, (row, col)), shape=(5,5)).todense()
        matrix([[ 0, 10, 20, 30, -1],
                [ 0,  0, 10, 20, -1],
                [ 0,  0,  0, 10, -1],
                [ 0,  0,  0,  0, -1],
                [ 0,  0,  0,  0,  0]])
        >>> chrom_list.tolist()
        ['a', 'a', 'a', 'a', '', 'a', 'a', 'a', '', 'a', 'a', '', 'a', '', 'b']
        """
        chrnamelist, startlist, endlist, extralist = zip(*cut_intervals)
        # now the distance between any two points
        # is computed and arranged such that for each
        # element of the data array, a corespondent distance is stored
        start_row = np.take(startlist, rows)
        start_col = np.take(startlist, cols)
        dist_list = start_col - start_row

        # now  all distances that are between chromosomes are removed
        # to do this I convert the array of chromosomes to
        # a array of indices. Then, when subtracting the
        # values that correspond to matrix.row and matrix.col
        # using the array of indices, any value other
        # than 0 means inter-chromosomal row,col combination.

        # chr_id_list is based on a trick using np.unique
        # to get from a list of strings
        # a list of integers
        chr_id_list = np.unique(chrnamelist, return_inverse=True)[1]

        chr_row = np.take(chr_id_list, rows)
        chr_col = np.take(chr_id_list, cols)
        chr_diff = chr_row - chr_col
        # set in dist_list array '-1' for all interchromosomal values
        dist_list[chr_diff != 0] = -1

        # make a corresponding chromosome name list
        # if filtering per chromosome is required
        chrom_list = np.take(chrnamelist, rows)
        chrom_list[chr_diff != 0] = ''

        return dist_list, chrom_list

    @staticmethod
    def fit_cut_intervals(cut_intervals):
        # check that the matrix has bins of same size
        # otherwise try to adjust the bins to
        # to match a regular binning
        chrom, start, end, extra = zip(*cut_intervals)

        median = int(np.median(np.diff(start)))
        diff = np.array(end) - np.array(start)
        # check if the bin size is homogeneous
        if len(np.flatnonzero(diff != median)) > (len(diff) * 0.01):
            # set the start position of a bin to the closest multiple
            # of the median
            def snap_nearest_multiple(start_x, m):
                resi = [-1 * (start_x % m), -start_x % m]
                return start_x + resi[np.argmin(np.abs(resi))]
            start = [snap_nearest_multiple(x, median) for x in start]
            end = [snap_nearest_multiple(x, median) for x in end]
            cut_intervals = zip(chrom, start, end, extra)
            sys.stderr.write('[getCountsByDistance] Bin size is not '
                             'homogeneous, setting \n'
                             'the bin distance to the median: {}\n'.format(median))
        return cut_intervals

    def convert_to_zscore_matrix(self, maxdepth=None, perchr=False):
        return self.convert_to_obs_exp_matrix(maxdepth=maxdepth, zscore=True, perchr=perchr)

    def convert_to_obs_exp_matrix(self, maxdepth=None, zscore=False, perchr=False):
        """
        Converts a corrected counts matrix into a
        obs / expected matrix or z-scores fast.

        The caveat is that the obs/exp or z-score are only
        computed for non-zero values, although zero values that
        are not part of the sparse matrix are considered.

        For each diagonal the mean (and std when computing z-scores) are
        calculated and then each non-zero value of the sparse matrix is
        replaced by the obs/exp or z-score.

        :param maxdepth:
        :param zscore: if a zscore wants to be returned instead of obs/exp
        :return: observed / expected sparse matrix

        from scipy.sparse import csr_matrix, dia_matrix
        >>> row, col = np.triu_indices(5)
        >>> cut_intervals = [('a', 0, 10, 1), ('a', 10, 20, 1),
        ... ('a', 20, 30, 1), ('a', 30, 40, 1), ('b', 40, 50, 1)]
        >>> hic = hiCMatrix()
        >>> hic.nan_bins = []
        >>> matrix = np.array([
        ... [ 1,  8,  5, 3, 0],
        ... [ 0,  4, 15, 5, 1],
        ... [ 0,  0,  0, 7, 2],
        ... [ 0,  0,  0, 0, 1],
        ... [ 0,  0,  0, 0, 0]])

        >>> hic.matrix = csr_matrix(matrix)
        >>> hic.setMatrix(hic.matrix, cut_intervals)
        >>> hic.convert_to_obs_exp_matrix().todense()
        matrix([[ 1. ,  0.8,  1. ,  1. ,  0. ],
                [ 0. ,  4. ,  1.5,  1. ,  1. ],
                [ 0. ,  0. ,  0. ,  0.7,  2. ],
                [ 0. ,  0. ,  0. ,  0. ,  1. ],
                [ 0. ,  0. ,  0. ,  0. ,  0. ]])

        >>> hic.matrix = csr_matrix(matrix)
        >>> hic.convert_to_obs_exp_matrix(maxdepth=20).todense()
        matrix([[ 1. ,  0.8,  1. ,  0. ,  0. ],
                [ 0. ,  4. ,  1.5,  1. ,  0. ],
                [ 0. ,  0. ,  0. ,  0.7,  nan],
                [ 0. ,  0. ,  0. ,  0. ,  nan],
                [ 0. ,  0. ,  0. ,  0. ,  0. ]])

        >>> hic.matrix = csr_matrix(matrix)
        >>> hic.convert_to_obs_exp_matrix(zscore=True).todense()
        matrix([[ 0.        , -0.56195149,         nan,         nan, -1.41421356],
                [ 0.        ,  1.93649167,  1.40487872,         nan,  0.        ],
                [ 0.        ,  0.        , -0.64549722, -0.84292723,  1.41421356],
                [ 0.        ,  0.        ,  0.        , -0.64549722,  0.        ],
                [ 0.        ,  0.        ,  0.        ,  0.        , -0.64549722]])

        nans occur where the standard deviation is zero
        """

        binsize = self.getBinSize()
        max_depth_in_bins = None

        if maxdepth:
            if maxdepth < binsize:
                exit("Please specify a maxDepth larger than bin size ({})".format(binsize))

            max_depth_in_bins = int(float(maxdepth * 1.5) / binsize)
            # work only with the upper matrix
            # and remove all pixels that are beyond
            # max_depth_in_bis
            # (this is done by subtracting a second sparse matrix
            # that contains only the upper matrix that wants to be removed.
            self.matrix = triu(self.matrix, k=0, format='csr') - \
                triu(self.matrix, k=max_depth_in_bins, format='csr')
        else:
            self.matrix = triu(self.matrix, k=0, format='csr')

        self.matrix.eliminate_zeros()
        depth = None
        if zscore is True:
            from scipy.sparse import diags
            m_size = self.matrix.shape[0]
            if max_depth_in_bins is not None:
                depth = max_depth_in_bins
            else:
                depth = m_size
                estimated_size_dense_matrix = m_size**2 * 8
                if estimated_size_dense_matrix > 100e6:
                    sys.stderr.write("To compute z-scores a dense matrix is required. This will use \n"
                                     "{} Mb of memory.\n To reduce memory use the maxdeph option."
                                     "".format(estimated_size_dense_matrix / 1e6))

            # to compute zscore the zero values need to be accounted and the matrix
            # need to become dense. This is only practical if only up to certain distance
            # wants to be evaluated, otherwise the dense matrix is too large.
            # To make the matrix dense and keep the same computations as when
            # the matrix is sparse the following is done:
            # A sparse diagonal matrix of shape = matrix.shape is created with ones
            # (only upper triangle contains diagonals up to maxdeph)
            # This  sparse matrix is then added to self.matrix
            # then, -1 is subtracted to the self.matrix.data, thus effectively
            # adding zeros.
            diag_mat_ones = diags(np.repeat([1], m_size * depth).reshape(depth, m_size), range(depth))

            self.matrix += diag_mat_ones

        from scipy.sparse import lil_matrix
        trasf_matrix = lil_matrix(self.matrix.shape)

        chr_submatrix = OrderedDict()
        cut_intervals = OrderedDict()
        chrom_sizes = OrderedDict()
        chrom_range = OrderedDict()
        if perchr:
            for chrname in self.getChrNames():
                chr_range = self.getChrBinRange(chrname)
                chr_submatrix[chrname] = self.matrix[chr_range[0]:chr_range[1], chr_range[0]:chr_range[1]].tocoo()
                cut_intervals[chrname] = [self.cut_intervals[x] for x in range(chr_range[0], chr_range[1])]
                chrom_sizes[chrname] = [chr_submatrix[chrname].shape[0]]
                chrom_range[chrname] = (chr_range[0], chr_range[1])

        else:
            chr_submatrix['all'] = self.matrix.tocoo()
            cut_intervals['all'] = self.cut_intervals
            chrom_sizes['all'] = np.array([v[1] - v[0] for k, v in self.chrBinBoundaries.iteritems()])
            chrom_range['all'] = (0, self.matrix.shape[0])

        for chrname, submatrix in chr_submatrix.iteritems():
            sys.stderr.write("processing chromosome {}\n".format(chrname))
            if zscore is True:
                # this step has to be done after tocoo()
                submatrix.data -= 1

            dist_list, chrom_list = self.getDistList(submatrix.row, submatrix.col,
                                                     hiCMatrix.fit_cut_intervals(cut_intervals[chrname]))

            # to get the sum of all values at a given distance I use np.bincount which
            # is quite fast. However, the input of bincount is positive integers. Moreover
            # it returns the sum for every consecutive integer, even if this is not on the list.
            # Thus, dist_list, which contains the distance in bp between any two bins is
            # converted to bin distance.

            # Because positive integers are needed we add +1 to all bin distances
            # such that the value of -1 (which means different chromosomes) can now be used

            dist_list[dist_list == -1] = -binsize
            # divide by binsize to get a list of bin distances and add +1 to remove negative values
            dist_list = (np.array(dist_list).astype(float) / binsize).astype(int) + 1

            # for each distance, return the sum of all values
            sum_counts = np.bincount(dist_list, weights=submatrix.data)
            distance_len = np.bincount(dist_list)
            # compute the average for each distance
            mat_size = submatrix.shape[0]
            mu = {}
            std = {}
            # compute mean value for each distance

            for bin_dist_plus_one, sum_value in enumerate(sum_counts):
                if maxdepth and bin_dist_plus_one == 0:  # this is for intra chromosomal counts
                    # when max depth is set, the computation
                    # of the total_intra is not accurate and is safer to
                    # output np.nan
                    mu[bin_dist_plus_one] = np.nan
                    std[bin_dist_plus_one] = np.nan
                    continue

                if bin_dist_plus_one == 0:
                    total_intra = mat_size ** 2 - sum([size ** 2 for size in chrom_sizes[chrname]])
                    diagonal_length = total_intra / 2
                else:
                    # to compute the average counts per distance we take the sum_counts and divide
                    # by the number of values on the respective diagonal
                    # which is equal to the size of each chromosome - the diagonal offset (for those
                    # chromosome larger than the offset)
                    # In the following example with two chromosomes
                    # the first (main) diagonal has a size equal to the matrix (6),
                    # while the next has 1 value less for each chromosome (4) and the last one has only 2 values

                    # 0 1 2 . . .
                    # - 0 1 . . .
                    # - - 0 . . .
                    # . . . 0 1 2
                    # . . . - 0 1
                    # . . . - - 0

                    # idx - 1 because earlier the values where
                    # shifted.
                    diagonal_length = sum([size - (bin_dist_plus_one - 1) for size in chrom_sizes[chrname] if size > (bin_dist_plus_one - 1)])

                # the diagonal length should contain the number of values at a certain distance.
                # If the matrix is dense, the distance_len[bin_dist_plus_one] correctly contains the number of values
                # If the matrix is equally spaced, then, the diagonal_length as computed before is accurate.
                # But, if the matrix is both sparse and with unequal bins, then none of the above methods is
                # accurate but the the diagonal_length as computed before will be closer.
                diagonal_length = max(diagonal_length, distance_len[bin_dist_plus_one])

                if diagonal_length == 0:
                    mu[bin_dist_plus_one] = np.nan
                else:
                    mu[bin_dist_plus_one] = np.float64(sum_value) / diagonal_length

                if np.isnan(sum_value):
                    sys.stderr.write("nan value found for distance {}\n".format((bin_dist_plus_one - 1) * binsize))

                # if zscore is needed, compute standard deviation: std = sqrt(mean(abs(x - x.mean())**2))
                if zscore:
                    values_sqrt_diff = \
                        np.abs((submatrix.data[dist_list == bin_dist_plus_one] - mu[bin_dist_plus_one])**2)
                    # the standard deviation is the sum of the differences with mu squared (value variable)
                    # plus all zeros that are not included in the sparse matrix
                    # for which the standard deviation is
                    # (0 - mu)**2 = (mu)**2
                    # The number of zeros is the diagonal length - the length of the non zero values
                    zero_values_sqrt_diff_sum = (diagonal_length - len(values_sqrt_diff)) * mu[bin_dist_plus_one]**2

                    _std = np.sqrt((values_sqrt_diff.sum() + zero_values_sqrt_diff_sum) / diagonal_length)
                    std[bin_dist_plus_one] = _std

            # use the expected values to compute obs/exp
            transf_ma = np.zeros(len(submatrix.data))
            for idx, value in enumerate(submatrix.data):
                if depth is not None and dist_list[idx] > depth + 1:
                    continue
                if zscore:
                    transf_ma[idx] = (value - mu[dist_list[idx]]) / std[dist_list[idx]]
                else:
                    transf_ma[idx] = value / mu[dist_list[idx]]
            submatrix.data = transf_ma
            trasf_matrix[chrom_range[chrname][0]:chrom_range[chrname][1], chrom_range[chrname][0]:chrom_range[chrname][1]] = submatrix.tolil()

        self.matrix = trasf_matrix.tocsr()

        return self.matrix

    def getCountsByDistance(self, mean=False, per_chr=False):
        """
        computes counts for each intrachromosomal distance.
        better used with a corrected matrix

        returns a dictionary having as key the distance
        and as value an array containing the matrix values
        corresponding to that distance

        >>> from scipy.sparse import coo_matrix
        >>> row, col = np.triu_indices(5)
        >>> cut_intervals = [('a', 0, 10, 1), ('a', 10, 20, 1),
        ... ('a', 20, 30, 1), ('a', 30, 40, 1), ('b', 40, 50, 1)]
        >>> hic = hiCMatrix()
        >>> hic.nan_bins = []
        >>> matrix = np.array([
        ... [ 0, 10,  5, 3, 0],
        ... [ 0,  0, 15, 5, 1],
        ... [ 0,  0,  0, 7, 3],
        ... [ 0,  0,  0, 0, 1],
        ... [ 0,  0,  0, 0, 0]])

        make the matrix symmetric:
        >>> hic.matrix = csr_matrix(matrix + matrix.T)
        >>> hic.setMatrix(csr_matrix(matrix + matrix.T), cut_intervals)
        >>> hic.getCountsByDistance()
        {0: array([0, 0, 0, 0, 0]), 10: array([10, 15,  7]), \
20: array([5, 5]), 30: array([3]), -1: array([0, 1, 3, 1])}

        Test get distance counts per chromosome
        >>> hic.distance_counts = None
        >>> hic.getCountsByDistance(per_chr=True)
        {'a': {0: array([0, 0, 0, 0]), 10: array([10, 15,  7]), \
20: array([5, 5]), 30: array([3])}, 'b': {0: array([0])}}

        Test the removal of masked bins
        >>> hic.nan_bins = [3]
        >>> hic.distance_counts = None
        >>> hic.getCountsByDistance()
        {0: array([0, 0, 0, 0]), 10: array([10, 15]), 20: array([5]), \
-1: array([0, 1, 3])}

        Test bins that are of different size
        >>> cut_intervals = [('a', 0, 12, 1), ('a', 12, 25, 1),
        ... ('a', 25, 35, 1), ('a', 35, 41, 1), ('b', 41, 50, 1)]
        >>> hic = hiCMatrix()
        >>> hic.nan_bins = []
        >>> hic.matrix = csr_matrix(matrix + matrix.T)
        >>> hic.setMatrix(csr_matrix(matrix + matrix.T), cut_intervals)
        >>> hic.getCountsByDistance()
        {0: array([0, 0, 0, 0, 0]), 33: array([3]), 11: array([10, 15,  7]), \
22: array([5, 5]), -1: array([0, 1, 3, 1])}
        """

        if self.distance_counts:
            return self.distance_counts

        cut_intervals = hiCMatrix.fit_cut_intervals(self.cut_intervals)

        M, N = self.matrix.shape
        # get the row and col indices of the upper triangle
        # of a square matrix of size M
        rows, cols = np.triu_indices(M)

        # put a negative number to all masked bins
        #
        # the upper triangle of the sparse data matrix is
        # converted into a dense matrix and represented
        # as a vector
        data = np.asarray(self.matrix.todense()[(rows, cols)]).flatten()

        # convert nans to zeros. Otherwise the computations will fail
        if np.any(np.isnan(data)):
            num_nan = len(np.flatnonzero(np.isnan(data)))
            sys.stderr.write("converting {} ({:.2f}) nans "
                             "to zeros".format(num_nan,
                                               float(num_nan) / len(data)))
            data[np.isnan(data)] = 0

        # get a vector of all distances. The order is the same
        # as the 'data' vector
        dist_list, chrom_list = hiCMatrix.getDistList(rows, cols,
                                                      cut_intervals)

        """
        remove all zeros corresponding to masked bins, needs to be done
        after the dist_list computation to get the distances right
        """
        # compute the overlap between rows and self.nan_bins
        # The row_res vector is boolean and contains 'False'
        # for all row ids that are in self.nan_bin
        row_res = np.invert(np.in1d(rows, self.nan_bins))
        col_res = np.invert(np.in1d(cols, self.nan_bins))

        # get data values for which row_res and col_res are both true
        data = data[row_res & col_res]
        dist_list = dist_list[row_res & col_res]
        chrom_list = chrom_list[row_res & col_res]

        if per_chr:
            distance = {}
            for chr_name in self.getChrNames():
                _chr_data = data[chrom_list == chr_name]
                _chr_dist_list = dist_list[chrom_list == chr_name]
                distance[chr_name] = hiCMatrix.dist_list_to_dict(_chr_data,
                                                                 _chr_dist_list)
        else:
            distance = hiCMatrix.dist_list_to_dict(data, dist_list)
        self.distance_counts = distance
        if mean:
            return [np.mean(distance[x]) for x in range(len(distance.keys()))]
        else:
            return distance

    @staticmethod
    def dist_list_to_dict(data, dist_list):
        """
        splits data, into numeric groups defined by dist_list
        Return a dictionary containing, for
        each unique distance a dictionary
        """

        order = np.argsort(dist_list)
        dist_list = dist_list[order]
        data = data[order]

        # having the dist_list sorted, np.split
        # is used to divide the data into
        # groups that lie at the same distance, for this
        # np.diff together with np.flatnonzero is used to
        # find the indices where the distance changes.
        # the '+1' is needed because the np.diff array is
        # one element smaller than the original array, thus
        # the indices based no the np.diff array are off by 1
        # with respect to the original array
        groups = np.split(data, np.flatnonzero(np.diff(dist_list)) + 1)

        # because the dist_list is sorted
        # the order of the unique values
        # corresponds to that of the groups.
        # In other words, group[0]
        # has distance_unique[0]
        # np.sort after np.unique  in theory
        # is not needed, but just in case...
        distance_unique = np.sort(np.unique(dist_list))

        # convert to dictionary having as key
        # the distance
        distance = {}
        for index in xrange(len(distance_unique)):
            distance[distance_unique[index]] = groups[index]

        return distance

    @staticmethod
    def getUnwantedChrs():
        # too drosophila specific. Should remove this function
        unwanted_chr = {'chrM', 'chrYHet', 'chrXHet', 'chrUextra', 'chrU', 'chr3RHet', 'chr3LHet', 'chr2RHet',
                        'chr2LHet'}
        return unwanted_chr

    def filterUnwantedChr(self, chromosome=None):
        if chromosome:
            self.keepOnlyTheseChr([chromosome])
        else:
            wanted = set(self.getChrNames()).difference(self.getUnwantedChrs())
            self.keepOnlyTheseChr(wanted)

    def keepOnlyTheseChr(self, chromosome_list):
        """
        given a list of chromosome names,
        these are kept, while any other is removed
        from the matrix
        """
        if isinstance(chromosome_list, str):
            chromosome_list = [chromosome_list]

        try:
            [self.chrBinBoundaries[x] for x in chromosome_list]
        except KeyError as e:
            raise Exception("Chromosome name {} not in matrix.".format(e))

        self.restoreMaskedBins()
        size = self.matrix.shape
        # initialize a 1D array containing the columns (and rows) to
        # select. By default none are selected
        sel = np.empty(size[0], dtype=np.bool)
        sel[:] = False

        for chrName in self.interval_trees.keys():
            if chrName not in chromosome_list:
                continue

            # identify start and end rows
            # of chromosomes that wants to be
            # kept
            index_start, index_end = self.getChrBinRange(chrName)
            sel[index_start:index_end] = True

        sel_id = np.flatnonzero(sel)
        mat = self.matrix[sel_id, :][:, sel_id]

        # update bin ids
        self.cut_intervals = [self.cut_intervals[x] for x in sel_id]

        # update correction factors
        if self.correction_factors is not None:
            self.correction_factors = [self.correction_factors[x] for x in sel_id]

        # keep track of nan bins
        if len(self.nan_bins):
            _temp = np.zeros(size[0])
            _temp[self.nan_bins] = 1
            _temp = _temp[sel_id]
            self.nan_bins = np.flatnonzero(_temp == 1)
        else:
            self.nan_bins = []

        self.numCols = len(sel_id)

        self.interval_trees, self.chrBinBoundaries = \
            self.intervalListToIntervalTree(self.cut_intervals)
        # remove distanceCounts
        try:
            self.distance_counts = None
        except AttributeError:
            pass
        self.matrix = mat
        return self.matrix

    def save_bing_ren(self, fileName):
        """
        Saves the matrix using bing ren's
        method which is chrom_name\tstart_bin\tend_bin\tvalues...
        """
        if os.path.isdir(fileName):
            exit('Please specify a file name to save the data. The given file name is a folder: \n{}\n'.format(fileName))

        if fileName[-3:] != '.gz':
            fileName += '.gz'

        try:
            fileh = gzip.open(fileName, 'w')
        except:
            msg = "{} file can't be opened for writing".format(fileName)
            raise msg

        colNames = []
        for x in range(self.matrix.shape[0]):
            chrom, start, end = self.cut_intervals[x][0:3]
            colNames.append("{}\t{}\t{}".format(chrom, start, end))

        for row in range(self.matrix.shape[0]):
            values = [str(x) for x in self.matrix[row, :].toarray().flatten()]
            fileh.write("{}\t{}\n".format(colNames[row], "\t".join(values)))

        fileh.close()

    def save_dekker(self, fileName):
        """
        Saves the matrix using dekker format
        """
        if os.path.isdir(fileName):
            exit('Please specify a file name to save the data. The given file name is a folder: \n{}\n'.format(fileName))

        if fileName[-3:] != '.gz':
            fileName += '.gz'

        try:
            fileh = gzip.open(fileName, 'w')
        except:
            msg = "{} file can't be opened for writing".format(fileName)
            raise msg

        colNames = []
        for x in range(self.matrix.shape[0]):
            chrom, start, end = self.cut_intervals[x][0:3]
            colNames.append("{}|--|{}:{}-{}".format(x, chrom, start, end))  # adds dm3 to the end (?problem..)

        fileh.write("#converted from hicexplorer\n")
        fileh.write("\t" + "\t".join(colNames) + "\n")
        for row in range(self.matrix.shape[0]):
            values = [str(x) for x in self.matrix[row, :].toarray().flatten()]
            fileh.write("{}\t{}\n".format(colNames[row], "\t".join(values)))

        fileh.close()

    def save_lieberman(self, fileName):
        """
        Saves the matrix using lieberman format. Given an output directory name and resolution of the matrix.

        In contrast to other methods, when saving using liebermans format a folder is required
        where the data is saved per chromosome
        """

        if os.path.isfile(fileName):
            exit("a file with the same name as the given path exists ({}). Please choose a folder.".format(fileName))

        if not os.path.isdir(fileName):
            os.mkdir(fileName)

        resolution = self.getBinSize()
        if not self.bin_size_homogeneous:
            sys.stderr.write("*WARNING* The hic matrix contains bins of difference size but\n"
                             "the 'lieberman' format requires equally spaced bins. The program\n"
                             "will proceed but the results may be unreliable.\n")

        for chrom in self.interval_trees.keys():
            chrstart, chrend = self.getChrBinRange(chrom)
            chrwise_mat = self.matrix[chrstart:chrend, chrstart:chrend]
            if len(chrwise_mat.data) > 0:
                sys.stderr.write("Saving chromosome {}...\n".format(chrom))
                chrwise_mat_coo = triu(chrwise_mat, k=0, format='csr').tocoo()
                start = chrwise_mat_coo.row * resolution
                end = chrwise_mat_coo.col * resolution
                fileh = gzip.open("{}/chr{}.gz".format(fileName, chrom), 'w')
                fileh.write("#converted from HiCExplorer format\n")
                for idx in range(len(start)):
                    fileh.write("{}\t{}\t{}\n".format(start[idx], end[idx], chrwise_mat_coo.data[idx]))
                fileh.close()

    def save_GInteractions(self, fileName):
        self.restoreMaskedBins()
        print self.matrix.shape
        mat_coo = triu(self.matrix, k=0, format='csr').tocoo()
        fileh = open("{}.tsv".format(fileName), 'w')
        for idx, counts in enumerate(mat_coo.data):
            chr_row, start_row, end_row, _ = self.cut_intervals[mat_coo.row[idx]]
            chr_col, start_col, end_col, _ = self.cut_intervals[mat_coo.col[idx]]
            fileh.write("{}\t{}\t{}\t{}\t{}\t{}\t{}\n".format(chr_row, int(start_row), int(end_row),
                                                              chr_col, int(start_col), int(end_col), counts))
        fileh.close()

    def save(self, filename):
        """
        Saves a matrix using hdf5 format
        :param filename:
        :return: None
        """
        self.restoreMaskedBins()
        if not filename.endswith(".h5"):
            filename += ".h5"

        # if the file name already exists
        # try to find a new suitable name
        if os.path.isfile(filename):
            sys.stderr.write("*WARNING* File already exists {}\n "
                             "Overwriting ...\n".format(filename))

            from os import unlink
            unlink(filename)
        try:
            nan_bins = np.array(self.nan_bins)
        except:
            nan_bins = np.array([])
        # save only the upper triangle of the
        # symmetric matrix
        matrix = triu(self.matrix, k=0, format='csr')
        with tables.open_file(filename, mode="w", title="HiCExplorer matrix") as h5file:
            matrix_group = h5file.create_group("/", "matrix", )
            # save the parts of the csr matrix
            for matrix_part in ('data', 'indices', 'indptr', 'shape'):
                arr = np.array(getattr(matrix, matrix_part))
                atom = tables.Atom.from_dtype(arr.dtype)
                ds = h5file.create_carray(matrix_group, matrix_part, atom, arr.shape)
                ds[:] = arr

            # save the matrix intervals
            intervals_group = h5file.create_group("/", "intervals", )
            chr_list, start_list, end_list, extra_list = zip(*self.cut_intervals)
            for interval_part in ('chr_list', 'start_list', 'end_list', 'extra_list'):
                arr = np.array(eval(interval_part))
                atom = tables.Atom.from_dtype(arr.dtype)
                ds = h5file.create_carray(intervals_group, interval_part, atom, arr.shape)
                ds[:] = arr

            # save nan bins
            if len(nan_bins):
                atom = tables.Atom.from_dtype(nan_bins.dtype)
                ds = h5file.create_carray(h5file.root, 'nan_bins', atom, nan_bins.shape)
                ds[:] = nan_bins

            # save corrections factors
            if self.correction_factors is not None and len(self.correction_factors):
                self.correction_factors = np.array(self.correction_factors)
                atom = tables.Atom.from_dtype(self.correction_factors.dtype)
                ds = h5file.create_carray(h5file.root, 'correction_factors', atom,
                                          self.correction_factors.shape)
                ds[:] = np.array(self.correction_factors)

            # save distance counts
            if self.distance_counts is not None and len(self.distance_counts):
                atom = tables.Atom.from_dtype(self.distance_counts.dtype)
                ds = h5file.create_carray(h5file.root, 'distance_counts', atom,
                                          self.distance_counts.shape)
                ds[:] = np.array(self.distance_counts)

    def save_npz(self, filename):
        """
        saves using the numpy npz format. Adequate for small samples.
        """
        self.restoreMaskedBins()
        chrNameList, startList, endList, extraList = zip(*self.cut_intervals)
        try:
            nan_bins = self.nan_bins
        except:
            nan_bins = np.array([])
        # save only the upper triangle of the
        # symmetric matrix
        matrix = triu(self.matrix, k=0, format='csr')
        try:
            np.savez(
                filename, matrix=matrix, chrNameList=chrNameList,
                startList=startList, endList=endList, extraList=extraList,
                nan_bins=nan_bins, correction_factors=self.correction_factors)
        except Exception as e:
            print "error saving matrix: {}".format(e)
            try:
                print "Matrix can not be saved because is too big!"
                print "Eliminating entries with only one count."

                # try to remove noise by deleting 1
                matrix.data = matrix.data - 1
                matrix.eliminate_zeros()
                np.savez(
                    filename, matrix=matrix, chrNameList=chrNameList,
                    startList=startList, endList=endList, extraList=extraList,
                    nan_bins=nan_bins)
            except:
                print "Matrix can not be saved because is too big!"
            exit()

    def diagflat(self, value=np.nan):
        """
        sets
        the matrix diagonal to np.nan
        """
        M = self.matrix.shape[0]
        diagmatrix = dia_matrix((np.repeat(value, M), 0), shape=(M, M))
        self_diag = dia_matrix(([self.matrix.diagonal()], [0]), shape=(M, M))
        # take matrix, subtract the values of the diagonal such that
        # it becomes all zeros, replace with new values by adding them
        self.matrix = self.matrix - self_diag + diagmatrix
        return self.matrix

    def filterOutInterChrCounts(self):
        """
        set all inter chromosomal counts to np.nan
        >>> from scipy.sparse import coo_matrix
        >>> row, col = np.triu_indices(5)
        >>> cut_intervals = [('a', 0, 10, 1), ('a', 10, 20, 1),
        ... ('a', 20, 30, 1), ('b', 30, 40, 1), ('b', 40, 50, 1)]
        >>> hic = hiCMatrix()
        >>> hic.nan_bins = []
        >>> matrix = np.array([
        ... [ 0, 10,  5, 3, 0],
        ... [ 0,  0, 15, 5, 1],
        ... [ 0,  0,  0, 7, 3],
        ... [ 0,  0,  0, 0, 1],
        ... [ 0,  0,  0, 0, 0]])

        make the matrix symmetric:
        >>> hic.matrix = csr_matrix(matrix + matrix.T)
        >>> hic.setMatrix(csr_matrix(matrix + matrix.T), cut_intervals)
        >>> hic.filterOutInterChrCounts().todense()
        matrix([[ 0, 10,  5,  0,  0],
                [10,  0, 15,  0,  0],
                [ 5, 15,  0,  0,  0],
                [ 0,  0,  0,  0,  1],
                [ 0,  0,  0,  1,  0]])
        """

        ma_coo = self.matrix.tocoo()
        dist_list, _ = hiCMatrix.getDistList(ma_coo.row, ma_coo.col,
                                             self.cut_intervals)

        # set to zero all cases in which dist_list is zero
        ma_coo.data[dist_list == -1] = 0

        self.matrix = ma_coo.tocsr()
        self.matrix.eliminate_zeros()
        return self.matrix

    def setMatrixValues(self, newMatrix):
        """
        replace the current matrix values
        by the given matrix values. The
        shapes have to coincide
        """
        assert self.matrix.shape == newMatrix.shape,\
            "Given matrix has different shape. New " \
            "values need to have the same shape as previous matrix."

        self.matrix = csr_matrix(newMatrix)

    def setCorrectionFactors(self, correction_factors):
        assert len(correction_factors) == self.matrix.shape[0], \
            "length of correction factors and length of matrix are different."
        self.correction_factors = correction_factors

    def reorderChromosomes_old(self, new_chr_order):
        if len(new_chr_order) != len(self.chrBinBoundaries):
            return
        dest = 0
        for chrName in new_chr_order:
            orig = self.chrBinBoundaries[chrName]

            self.reorderMatrix(orig, dest)
            dest += orig[1] - orig[0]

    def reorderChromosomes(self, new_chr_order):
        new_order = []
        for chrName in new_chr_order:
            # check that the chromosome names are valid
            if chrName not in self.chrBinBoundaries:
                exit("Chromosome name '{}' not found. Please check the correct spelling "
                     "of the chromosomes and try again".format(chrName))
            orig = self.chrBinBoundaries[chrName]
            new_order.extend(range(orig[0], orig[1]))
        self.reorderBins(new_order)

    def reorderBins(self, new_order):
        """
        reorders the rows and colums of the
        matrix according to the new order.
        The new order can be smaller
        than the original matrix. In that
        case, the ids not in the
        new order are removed.
        """
        orig_num_rows = self.matrix.shape[0]
        self.matrix = self.matrix[new_order, :][:, new_order]
        self.cut_intervals = [self.cut_intervals[x] for x in new_order]
        # reorder the masked bins
        # keep track of nan bins
        if len(self.nan_bins):
            _temp = np.zeros(orig_num_rows)
            _temp[self.nan_bins] = 1
            _temp = _temp[new_order]
            self.nan_bins = np.flatnonzero(_temp == 1)
        else:
            self.nan_bins = []

        self.interval_trees, self.chrBinBoundaries = \
            self.intervalListToIntervalTree(self.cut_intervals)

    def removeBins(self, bin_ids):
        """ given an array of ids, all rows and columns
        matching those ids are removed
        """
        rows = cols = np.delete(range(self.matrix.shape[1]), bin_ids)

        self.matrix = self.matrix[rows, :][:, cols]
        self.cut_intervals = [self.cut_intervals[x] for x in rows]
        self.interval_trees, self.chrBinBoundaries = \
            self.intervalListToIntervalTree(self.cut_intervals)

    def maskBins(self, bin_ids):
        """
        Mask the list of bins given. Mask means
        to remove the bins from the matrix,
        and keep the information about the intervals
        as masked
        """
        if len(bin_ids) == 0:
            return

        self.printchrtoremove(bin_ids, restore_masked_bins=False)
        try:
            # check if a masked bin already exists
            if len(self.orig_bin_ids) > 0:
                print "Masked bins already present"
                M = self.matrix.shape[0]
                previous_bin_ids = self.orig_bin_ids[M:]
                # merge new and old masked bins
                bin_ids = np.unique(np.concatenate([previous_bin_ids, self.orig_bin_ids[bin_ids]]))
                np.sort(bin_ids)
                self.restoreMaskedBins()
        except:
            pass

        # join with existing nan_bins
        if len(self.nan_bins) > 0:
            print "found existing {} nan bins that will be " \
                "included for masking ".format(len(self.nan_bins))
            bin_ids = np.unique(np.concatenate([self.nan_bins, bin_ids]))
            self.nan_bins = []
        rows = cols = np.delete(range(self.matrix.shape[1]), bin_ids)
        self.matrix = self.matrix[rows, :][:, cols]

        # to keep track of removed bins
        # I add their ids to the end of the rows vector
        # to reverse the changes, I just need to do an argsort
        # to put the removed bins in place
        self.orig_bin_ids = np.concatenate([rows, bin_ids])

        new_cut_intervals = [self.cut_intervals[x] for x in rows]

        self.orig_cut_intervals = \
            new_cut_intervals + [self.cut_intervals[x] for x in bin_ids]

        self.cut_intervals = new_cut_intervals
#        self.nan_bins = np.intersect1d(self.nan_bins, rows)
        self.interval_trees, self.chrBinBoundaries = \
            self.intervalListToIntervalTree(self.cut_intervals)

        if self.correction_factors is not None:
            self.correction_factors = self.correction_factors[rows]

    def update_matrix(self, new_matrix, new_cut_intervals):
        """
        give a new matrix and list of cut intervals, the matrix,  cut intervals and
        the respective tree are updated
        :param new_matrix: now values for the sparse matrix
        :param new_cut_intervals: list of cut intervals, each entry being a tuple of the form
        (chrom, start, end, coverage)
        :return:
        """
        if hasattr(self, 'orig_bin_ids'):
            raise Exception("matrix contains masked bins. Restore masked bins first")

        assert len(new_cut_intervals) == new_matrix.shape[0], "matrix shape and len of cut intervals do not match"

        self.matrix = new_matrix
        self.cut_intervals = new_cut_intervals

        self.interval_trees, self.chrBinBoundaries = \
            self.intervalListToIntervalTree(self.cut_intervals)

        self.nan_bins = np.flatnonzero(self.matrix.sum(0).A == 0)

    def restoreMaskedBins(self):
        """
        Puts backs into the matrix the bins
        removed
        """
        try:
            self.orig_bin_ids
        except AttributeError:
            return
        # the rows to add are
        # as an empty sparse matrix
        M = self.matrix.shape[0]
        N = len(self.orig_bin_ids) - M
        rows_mat = csr_matrix((N, M))
        # cols to add
        cols_mat = csr_matrix((M + N, N))

        # add the rows and cols at the end of the
        # current matrix
        self.matrix = sparse_vstack([self.matrix, rows_mat])
        self.matrix = sparse_hstack([self.matrix, cols_mat], format='csr')

        # the new matrix has the right number of cols and rows, now
        # they need to be reordered to be back in their original places
        rows = cols = np.argsort(self.orig_bin_ids)
        self.matrix = self.matrix[rows, :][:, cols]
        self.cut_intervals = [self.orig_cut_intervals[x] for x in rows]
        self.interval_trees, self.chrBinBoundaries = \
            self.intervalListToIntervalTree(self.cut_intervals)
        # set as nan_bins the masked bins that were restored
        self.nan_bins = self.orig_bin_ids[M:]

        if self.correction_factors is not None:
            # add missing values as nans at end of array
            self.correction_factors = np.concatenate([self.correction_factors,
                                                      np.repeat(np.nan, N)])
            # reorder array
            self.correction_factors = self.correction_factors[rows]
        del(self.orig_cut_intervals, self.orig_bin_ids)
        sys.stderr.write("masked bins were restored\n")

    def reorderMatrix(self, orig, dest):
        """
        Given a matrix, a region over the diagonal is moved from
        its origin to a new destination. With this method a
        new order of the chromosomes can be produced.
        :param orig: a tuple containing the indices of the region to be moved
        :param dest: the index of the region into which to insert
                     the section moved
        """

        rows = np.delete(range(self.matrix.shape[1]), range(orig[0], orig[1]))

        if dest > orig[1]:
            dest = dest - (orig[1] - orig[0])

        rows = cols = np.insert(
            rows, np.repeat(dest, orig[1] - orig[0]), range(orig[0], orig[1]))
        self.matrix = self.matrix[rows, :][:, cols]
        self.cut_intervals = [self.cut_intervals[x] for x in rows]
        self.interval_trees, self.chrBinBoundaries = \
            self.intervalListToIntervalTree(self.cut_intervals)

        if self.correction_factors is not None:
            self.correction_factors = self.correction_factors[rows]
        return

    def truncTrans(self, high=0.05):
        """Truncates trans contacts to remove blowouts
        Clip high counts in trans regions (i.e. between
        chromosomes) to the max value found in the 1-high*100
        percentile

        :param:  high : float, 0<high<1, optional
            Fraction of top trans interactions to be removed
        """
        mat = self.matrix.tocoo()
        dist_list = hiCMatrix.getDistList(mat.row, mat.col, self.cut_intervals)
        if np.count_nonzero(dist_list == -1) == 0:
            return
        max_inter = np.percentile(mat.data[dist_list == -1], (100 - high))
        mat.data[(mat.data >= max_inter) & (dist_list == -1)] == max_inter

        self.setMatrixValues(mat)

    def truncTrans_bk(self, high=0.0005):
        """Truncates trans contacts to remove blowouts

        :param:  high : float, 0<high<1, optional
            Fraction of top trans interactions to be removed
        """
        M, N = self.matrix.shape
        rows, cols = np.triu_indices(M, 1)
        dist_list = hiCMatrix.getDistList(rows, cols)
        data = np.asarray(self.matrix.todense()[(rows, cols)]).flatten()

        max_inter = np.percentile(data[dist_list == -1], 100. * (1 - high))

        to_clip = np.flatnonzero((data >= max_inter) & (dist_list == -1))
        rows_to_clip = rows[to_clip]
        cols_to_clip = cols[to_clip]
        mat = self.matrix.tolil()
        mat[(rows_to_clip, cols_to_clip)] = max_inter
        mat[(cols_to_clip, rows_to_clip)] = max_inter

        self.setMatrixValues(mat)

    def removePoorRegions(self, cutoff=2):
        """Removes "cutoff" percent of bins with least counts.
        Also, removes bins that have less than sequencedFraction
        unique reads  with respect
        to the total number of reads found in the bin. (which
        includes multi reads).

        :param:  cutoff : int, 0<cutoff<100
        Percent of lowest-counts bins to be removed

        """
        # remove all nan values
        self.matrix.data[np.isnan(self.matrix.data)] = 0
        # using the row sum, find and remove those
        # bins that have the botton 1%
        row_sum = np.asarray(self.matrix.sum(axis=1)).flatten()
        min_threshold = np.percentile(row_sum, cutoff)
        to_remove = np.flatnonzero(row_sum <= min_threshold)
        # code to count were the bins are going to be removed
        chrom, start, end, cov = zip(*self.cut_intervals)
        chrom = np.sort(np.array(chrom)[to_remove])
        chr_count = np.diff(np.concatenate(
            [np.unique(np.sort(chrom), return_index=True)[1],
             [len(chrom)]]))
        chr_dict = dict(zip(chrom[chr_count], chr_count))
        sys.stderr.write('num poor regions to remove {}\n{}\n'.format(
            len(to_remove),
            chr_dict))
        self.maskBins(to_remove)

    def printchrtoremove(self, to_remove, label="Number of poor regions to remove", restore_masked_bins=True):
        """
        prints out the number of bin per chromosomes
        that will be removed
        """
        cnt = {}
        try:
            self.prev_to_remove
        except:
            self.prev_to_remove = np.array([])

        # if the same information was already printed don't
        # show it again.
        if np.array_equal(self.prev_to_remove, to_remove):
            return

        if restore_masked_bins:
            try:
                # check if a masked bin already exists
                if len(self.orig_bin_ids) > 0:
                    print "Masked bins already present"
                    self.restoreMaskedBins()
            except:
                pass
        for idx in to_remove:
            chrom = self.cut_intervals[idx][0]
            if chrom not in cnt:
                cnt[chrom] = 0
            cnt[chrom] += 1

        sys.stderr.write('{}: {}\n{}\n'.format(label, len(to_remove), cnt))
        self.prev_to_remove = to_remove

    def removeBySequencedCount(self, sequencedFraction=0.5):
        """
        Removes bins that have less than sequencedFraction
        unique reads  with respect
        to the total number of reads found in the bin. (which
        includes multi reads).

        :param:   sequencedFraction: float, optional, 0<x<1
            Fraction of the bin that needs to be sequenced in order
            to keep the bin

        """
        chrNameList, startList, endList, coverage = zip(*self.cut_intervals)
        if type(coverage[0]) != np.float64:
            return
        to_remove = np.flatnonzero(np.array(coverage) < sequencedFraction)
        self.maskBins(to_remove)
        return to_remove

    def get_chromosome_sizes(self):
        chrom_sizes = OrderedDict()
        for chrom, (start_bin, end_bin) in self.chrBinBoundaries.iteritems():
            chrom, start, end, _ = self.cut_intervals[end_bin - 1]
            chrom_sizes[chrom] = end
        return chrom_sizes

    @staticmethod
    def intervalListToIntervalTree(interval_list):
        """
        given an ordered list of (chromosome name, start, end)
        this is transformed to a number of interval trees,
        one for each chromosome
        """

        assert len(interval_list) > 0, "List is empty"
        cut_int_tree = {}
        chrbin_boundaries = OrderedDict()
        intval_id = 0
        chr_start_id = 0
        previous_chrom = None
        for intval in interval_list:
            chrom, start, end = intval[0:3]
            start = int(start)
            end = int(end)
            if previous_chrom != chrom:
                if previous_chrom is None:
                    previous_chrom = chrom

                chrbin_boundaries[previous_chrom] = \
                    (chr_start_id, intval_id)
                chr_start_id = intval_id
                cut_int_tree[chrom] = IntervalTree()
                previous_chrom = chrom

            cut_int_tree[chrom].add(Interval(start, end, intval_id))

            intval_id += 1

        chrbin_boundaries[chrom] = (chr_start_id, intval_id)

        return cut_int_tree, chrbin_boundaries



import argparse


def getParentArgParse(args=None):
    parser = argparse.ArgumentParser(add_help=False)

    parser.add_argument('--matrix', '-m',
                        help='path of the  Hi-C matrix.',
                        required=True)

    return parser


def writableFile(string):
    try:
        open(string, 'w').close()
    except:
        msg = "{} file can be opened for writting".format(string)
        raise argparse.ArgumentTypeError(msg)
    return string



import multiprocessing
import numpy as np

def mapReduce(staticArgs, func, list_contig_sizes,
              contig_list_length=None, 
              region=None,
              numberOfProcessors=4,
              verbose=False):

    """
    Split the genome into parts that are sent to workers using a defined
    number of procesors. Results are collected and returned.

    For each genomic region the given 'func' is called using
    the following parameters:

     contigList, staticArgs

    The *arg* are static, *pickable* variables that need to be sent
    to workers.

    The contigList length corresponds to the number of
    contigs that should be processed by each worker.

    Depending on the type of process a larger or shorter regions may be 
    preferred

    :param list_contig_sizes: A list of duples containing the contig id and its length
    """

    if not contig_list_length:
        contig_list_length = len(list_contig_sizes) / numberOfProcessors
        
    if verbose:
        print "contigs per worker: %s" % (contig_list_length) 

    TASKS = []
    # iterate over all contigs
    contig_list, size = zip(*list_contig_sizes)
    contig_blocks = np.array_split(contig_list, numberOfProcessors)
    for contig_list in contig_blocks:
        argsList = [contig_list.tolist()]
        argsList.extend(staticArgs)
        TASKS.append( tuple(argsList) )

    if len(TASKS) > 1 and numberOfProcessors>1:
        if verbose:
            print ("using {} processors for {} "
                   "number of tasks".format(numberOfProcessors,
                                            len(TASKS)))

        pool = multiprocessing.Pool(numberOfProcessors)
        res = pool.map_async( func, TASKS ).get(9999999)
    else:
        res = map(func, TASKS)

    return res



import sys
import argparse
from scipy.sparse import lil_matrix
import logging

from hicexplorer.iterativeCorrection import iterativeCorrection
from hicexplorer import HiCMatrix as hm
from hicexplorer._version import __version__

import numpy as np
debug = 0
logging.basicConfig()
log = logging.getLogger("hicCorrectMatrix")
log.setLevel(logging.WARN)


def parse_arguments(args=None):

    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        conflict_handler='resolve',
        description="""
Iterative correction for a hic matrix (see Imakaev et al. 2012
Nature Methods for details). For the method to work correctly, bins
with low or too high coverage need to be filtered. For this, it is
recommended to first run some diagnostic plots to determine the
modified z-score cut off.

It is recommended to run hicCorrectMatrix as follows:

    $ hicCorrectMatrix diagnostic_plot --matrix hic_matrix -o plot_file.png

Then, after revising the plot and deciding the threshold values:

    $ hicCorrectMatrix correct --matrix hic_matrix \\
         --filterThreshold <lower threshold> <upper threshold> -o corrected_matrix

"""
    )

    parser.add_argument('--version', action='version',
                        version='%(prog)s {}'.format(__version__))

    subparsers = parser.add_subparsers(
        title="commands",
        dest='command',
        metavar='',
        help="""
        To get detailed help on each of the options:

            $ hicCorrectMatrix diagnostic_plot -h

            $ hicCorrectMatrix correct -h
            """)
    subparsers.add_parser(
        'correct',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        parents=[correct_subparser()],
        help="Run the iterative correction.",
        usage='%(prog)s '
              '--matrix hic_matrix.h5 '
              '--filterThreshold -1.2 5'
              '-out corrected_matrix.h5 \n')

    plot_mode = subparsers.add_parser(
        'diagnostic_plot',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        help="Plots a histogram of the coverage per bin together with the modified z-score "
             "based on the median absolute deviation method (see Boris Iglewicz and David Hoaglin 1993, "
             "Volume 16: How to Detect and Handle Outliers The ASQC Basic References in Quality Control:  "
             "Statistical Techniques, Edward F. Mykytka, Ph.D., Editor. ",
        usage='%(prog)s '
              '--matrix hic_matrix.h5 '
              '-o file.png')
    plot_mode.add_argument('--matrix', '-m',
                           help='Hi-C matrix.',
                           required=True)

    plot_mode.add_argument('--plotName', '-o',
                           help='File name to save the diagnostic plot.',
                           required=True)

    plot_mode.add_argument('--chromosomes',
                           help='List of chromosomes to be included in the iterative '
                           'correction. The order of the given chromosomes will be then '
                           'kept for the resulting corrected matrix',
                           default=None,
                           nargs='+')

    plot_mode.add_argument('--xMax',
                           help='Max value for the x-axis in counts per bin',
                           default=None,
                           type=float)

    plot_mode.add_argument('--perchr',
                           help='Compute histogram per chromosome. For samples from cells with uneven number '
                                'of chromosomes and/or translocations it is advisable to check the histograms '
                                'per chromosome to find the most conservative `filterThreshold`.',
                           action='store_true')

    parser.add_argument('--verbose',
                        help='Print processing status',
                        action='store_true')

    return parser


def correct_subparser():
    # define the arguments
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument('--matrix', '-m',
                        help='Hi-C matrix.',
                        required=True)

    parser.add_argument('--iterNum', '-n',
                        help='number of iterations',
                        type=int,
                        metavar='INT',
                        default=500)

    parser.add_argument('--outFileName', '-o',
                        help='File name to save the resulting matrix. The '
                             'output is a .h5 file.',
                        required=True)

    parser.add_argument('--filterThreshold', '-t',
                        help='Bins of low coverage or large coverage need to be removed. '
                             'Usually they do not contain valid Hi-C data of represent '
                             'regions that accumulate reads. Use hicCorrectMatrix diagnostic_plot '
                             'to identify the modified z-value thresholds. A lower and upper '
                             'threshold are required separated by space. Eg. --filterThreshold '
                             '-1.5 5',
                        type=float,
                        nargs=2,
                        required=True)

    parser.add_argument('--inflationCutoff',
                        help='Value corresponding to the maximum number of times a bin '
                        'can be scaled up during the iterative correction. For example, '
                        'a inflation Cutoff of 3 will filter out all bins that were '
                        'expanded 3 times or more during the iterative correction.',
                        type=float)

    parser.add_argument('--transCutoff', '-transcut',
                        help='Clip high counts in the top -transcut trans '
                        'regions (i.e. between chromosomes). A usual value '
                        'is 0.05 ',
                        type=float)

    parser.add_argument('--sequencedCountCutoff',
                        help='Each bin receives a value indicating the '
                        'fraction that is covered by reads. A cutoff of '
                        '0.5 will discard all those bins that have less '
                        'than half of the bin covered.',
                        default=None,
                        type=float)

    parser.add_argument('--chromosomes',
                        help='List of chromosomes to be included in the iterative '
                        'correction. The order of the given chromosomes will be then '
                        'kept for the resulting corrected matrix',
                        default=None,
                        nargs='+')

    parser.add_argument('--skipDiagonal', '-s',
                        help='If set, diagonal counts are not included',
                        action='store_true')

    parser.add_argument('--perchr',
                        help='Normalize each chromosome separately',
                        action='store_true')

    parser.add_argument('--verbose',
                        help='Print processing status',
                        action='store_true')
    parser.add_argument('--version', action='version',
                        version='%(prog)s {}'.format(__version__))

    return parser


def iterative_correction(matrix, args):
    corrected_matrix, correction_factors = iterativeCorrection(matrix,
                                                               M=args.iterNum,
                                                               verbose=args.verbose)

    return corrected_matrix, correction_factors


def fill_gaps(hic_ma, failed_bins, fill_contiguous=False):
    """ try to fill-in the failed_bins the matrix by adding the
    average values of the neighboring rows and cols. The idea
    for the iterative correction is that is best to put
    something in contrast to not put anything

    hic_ma: hic matrix object
    failed_bins: list of bin ids
    fill_contiguous: If True, stretches of masked rows/cols are filled.
                     Otherwise, these cases are skipped

    """
    log.info("starting fill gaps")
    mat_size = hic_ma.matrix.shape[0]
    fill_ma = hic_ma.matrix.copy().tolil()
    if fill_contiguous is True:
        discontinuous_failed = failed_bins
        consecutive_failed_idx = np.array([])
    else:
        # find stretches of consecutive failed regions
        consecutive_failed_idx = np.flatnonzero(np.diff(failed_bins) == 1)
        # the banned list of indices is equal to the actual list
        # and the list plus one, to identify consecutive failed regions.
        # for [1,2,5,10] the np.diff is [1,3,5]. The consecutive id list
        # is [0], for '1', in the original list, but we are missing the '2'
        # thats where the consecutive_failed_idx+1 comes.
        consecutive_failed_idx = np.unique(np.sort(
            np.concatenate([consecutive_failed_idx,
                            consecutive_failed_idx + 1])))
        # find the failed regions that are not consecutive
        discontinuous_failed = [x for idx, x in enumerate(failed_bins)
                                if idx not in consecutive_failed_idx]

    sys.stderr.write("Filling {} failed bins\n".format(
        len(discontinuous_failed)))

    """
    for missing_bin in discontinuous_failed:
        if 0 < missing_bin < mat_size - 1:
            for idx in range(1, mat_size - 2):
                if idx % 100 == 0:
                    sys.stderr.write(".")
                # the new row value is the mean between the upper
                # and lower bins corresponding to the same diagonal
                fill_ma[missing_bin, idx :] = \
                    (hic_ma.matrix[missing_bin-1, idx-1] +
                     hic_ma.matrix[missing_bin+1, idx+1]) / 2

                # same for cols
                fill_ma[idx, missing_bin] = \
                    (hic_ma.matrix[idx-1, missing_bin-1] +
                     hic_ma.matrix[idx+1, missing_bin+1]) / 2

    """
    for missing_bin in discontinuous_failed:
        if 0 < missing_bin < mat_size - 1:
            # the new row value is the mean between the upper
            # and lower rows
            fill_ma[missing_bin, 1:mat_size - 1] = \
                (hic_ma.matrix[missing_bin - 1, :mat_size - 2] +
                 hic_ma.matrix[missing_bin + 1, 2:]) / 2

            # same for cols
            fill_ma[1:mat_size - 1, missing_bin] = \
                (hic_ma.matrix[:mat_size - 2, missing_bin - 1] +
                 hic_ma.matrix[2:, missing_bin + 1]) / 2

    # identify the intersection points of the failed regions because they
    # neighbors get wrong values
    for bin_a in discontinuous_failed:
        for bin_b in discontinuous_failed:
            if 0 < bin_a < mat_size and \
                    0 < bin_b < mat_size:
                # the fill value is the average over the
                # neighbors that do have a value

                fill_value = np.mean([
                    hic_ma.matrix[bin_a - 1, bin_b - 1],
                    hic_ma.matrix[bin_a - 1, bin_b + 1],
                    hic_ma.matrix[bin_a + 1, bin_b - 1],
                    hic_ma.matrix[bin_a + 1, bin_b + 1],
                ])

                fill_ma[bin_a - 1, bin_b] = fill_value
                fill_ma[bin_a + 1, bin_b] = fill_value
                fill_ma[bin_a, bin_b - 1] = fill_value
                fill_ma[bin_a, bin_b + 1] = fill_value

    # return the matrix and the bins that continue to be failed regions
    return fill_ma.tocsr(), np.sort(failed_bins[consecutive_failed_idx])


class MAD(object):

    def __init__(self, points):
        """
        Returns a boolean array with True if points are outliers and False
        otherwise.

        :param points: An array

        :returns: A numobservations-length boolean array.

        :references: Boris Iglewicz and David Hoaglin (1993), "Volume 16: How to Detect and
            Handle Outliers", The ASQC Basic References in Quality Control:
            Statistical Techniques, Edward F. Mykytka, Ph.D., Editor.
        """

        self.mad_b_value = 0.6745
        if len(points.shape) == 1:
            points = points[:, None]
        self.median = np.median(points[points > 0], axis=0)

        diff = np.sum((points - self.median), axis=-1)

        self.med_abs_deviation = np.median(np.abs(diff))
        self.modified_z_score = self.mad_b_value * diff / self.med_abs_deviation

    def get_motified_zscores(self):

        return self.modified_z_score

    def is_outlier(self, lower_threshold, upper_threshold):
        """
        Returns a boolean list of outliers

        :param lower_threshold: Lower median absolute deviation
        :param upper_threshold: upper median absolute deviation

        :return: boolean array
        """

        return (self.modified_z_score < lower_threshold) | \
               (self.modified_z_score > upper_threshold)

    def value_to_mad(self, value):
        """
        return the mad value for a given value
        based on the data
        """

        diff = value - self.median
        return self.mad_b_value * diff / self.med_abs_deviation

    def mad_to_value(self, mad):
        """
        return the numeric value for a given mad score
        based on the data

        z = b_v * (x - median) / mad
        z * mad / b_v = x - median
        (z * mad / b_v) + median = x
        """

        return (mad * self.med_abs_deviation / self.mad_b_value) + self.median


def plot_total_contact_dist(hic_ma, args):
    """
    Plots the distribution of number of contacts (excluding self contacts)
    Outliers with a high number are removed for the plot

    :param hic_ma: sparse matrix
    :return:
    """
    from matplotlib import use
    use('Agg')
    import matplotlib.pyplot as plt
    import matplotlib.gridspec as gridspec
    from matplotlib.ticker import MultipleLocator, FormatStrFormatter
    majorlocator = MultipleLocator(1)
    majorformatter = FormatStrFormatter('%d')
    minorlocator = MultipleLocator(0.2)

    def plot_histogram(row_sum_values, mad_values, ax1, title=None):

        if args.xMax:
            ax1.set_xlim(ax1.get_xlim()[0], args.xMax)
            row_sum_values = row_sum_values[row_sum_values < args.xMax]

        ax1.set_xlabel("total counts per bin")
        ax1.set_ylabel("frequency")
    #    ax1.xaxis.grid(True)
        ax1.patch.set_visible(False)
        dist, bin_s, __ = ax1.hist(row_sum_values, 100, color='green')

        # add second axis on top
        ax2 = ax1.twiny()
        ax2.set_xlabel("modified z-score")
        ax2.xaxis.set_major_locator(majorlocator)
        ax2.xaxis.set_major_formatter(majorformatter)
        ax2.xaxis.grid(True, which='minor')
        # for the minor ticks, use no labels; default NullFormatter
        ax2.xaxis.set_minor_locator(minorlocator)

        # update second axis values by mapping the min max
        # of the main axis to the translated values
        # into modified z score.
        ax2.set_xlim(mad_values.value_to_mad(np.array(ax1.get_xlim())))

        # get first local mininum value
        local_min = [x for x, y in enumerate(dist) if 1 <= x < len(dist) - 1 and
                     dist[x - 1] > y < dist[x + 1]]

        if len(local_min) > 0:
            threshold = bin_s[local_min[0]]
        else:
            threshold = None

        if threshold:
            mad_threshold = mad_values.value_to_mad(threshold)
            ymin, ymax = ax2.get_ylim()
            ax2.vlines(mad_threshold, ymin, ymax)
            if title:
                print "{}: mad threshold {}".format(title, mad_threshold)
            else:
                print "mad threshold {}".format(mad_threshold)

    # replace nan by 0
    hic_ma.matrix.data[np.isnan(hic_ma.matrix.data)] = 0

    if args.perchr:
        chroms = hic_ma.getChrNames()
        if len(chroms) > 30:
            sys.stderr.write("The matrix contains {} chromosomes. It is not practical to plot "
                             "each. Try using --chromosomes to select some chromosomes or "
                             "plot a single histogram.")
        num_rows = int(np.ceil(float(len(chroms)) / 5))
        num_cols = min(len(chroms), 5)
        grids = gridspec.GridSpec(num_rows, num_cols)
        fig = plt.figure(figsize=(6 * num_cols, 5 * num_rows))
        ax = {}
        for plot_num, chrname in enumerate(chroms):
            log.info("Plotting chromosome {}".format(chrname))

            chr_range = hic_ma.getChrBinRange(chrname)
            chr_submatrix = hic_ma.matrix[chr_range[0]:chr_range[1], chr_range[0]:chr_range[1]]

            row_sum = np.asarray(chr_submatrix.sum(axis=1)).flatten()
            row_sum = row_sum - chr_submatrix.diagonal()
            mad = MAD(row_sum)
            modified_z_score = mad.get_motified_zscores()

            # high remove outliers
            row_sum = row_sum[modified_z_score < 5]

            col = plot_num % num_cols
            row = int(float(plot_num) / num_cols)
            ax[chrname] = fig.add_subplot(grids[row, col])

            plot_histogram(row_sum, mad, ax[chrname], title=chrname)
            ax[chrname].set_title(chrname)
    else:
        fig = plt.figure()
        row_sum = np.asarray(hic_ma.matrix.sum(axis=1)).flatten()
        row_sum = row_sum - hic_ma.matrix.diagonal()
        mad = MAD(row_sum)
        modified_z_score = mad.get_motified_zscores()

        # high remove outliers
        row_sum = row_sum[modified_z_score < 5]
        ax = fig.add_subplot(111)
        plot_histogram(row_sum, mad, ax)

    plt.tight_layout()
    plt.savefig(args.plotName)
    plt.close()


def filter_by_zscore(hic_ma, lower_threshold, upper_threshold, perchr=False):
    """
    The method defines thresholds per chromosome
    to avoid introducing bias due to different chromosome numbers

    """
    to_remove = []
    if perchr:
        for chrname in hic_ma.interval_trees.keys():
            chr_range = hic_ma.getChrBinRange(chrname)
            chr_submatrix = hic_ma.matrix[chr_range[0]:chr_range[1],
                                          chr_range[0]:chr_range[1]]

            # replace nan values by zero
            chr_submatrix.data[np.isnan(chr_submatrix.data)] = 0
            row_sum = np.asarray(chr_submatrix.sum(axis=1)).flatten()
            # subtract from row sum, the diagonal
            # to account for interactions with other bins
            # and not only self interactions that are the dominant count
            row_sum = row_sum - chr_submatrix.diagonal()
            mad = MAD(row_sum)
            problematic = np.flatnonzero(mad.is_outlier(lower_threshold, upper_threshold))

            # because the problematic indices are specific for the given chromosome
            # they need to be updated to match the large matrix indices
            problematic += chr_range[0]

            if len(problematic) == 0:
                log.warn("Warning. No bins removed for chromosome {} using thresholds {} {}"
                         "\n".format(chrname, lower_threshold, upper_threshold))

            to_remove.extend(problematic)
    else:
        row_sum = np.asarray(hic_ma.matrix.sum(axis=1)).flatten()
        # subtract from row sum, the diagonal
        # to account for interactions with other bins
        # and not only self interactions that are the dominant count
        row_sum = row_sum - hic_ma.matrix.diagonal()
        mad = MAD(row_sum)
        to_remove = np.flatnonzero(mad.is_outlier(lower_threshold, upper_threshold))

    return sorted(to_remove)


def main():
    args = parse_arguments().parse_args()
    if args.verbose:
        log.setLevel(logging.INFO)

    ma = hm.hiCMatrix(args.matrix)

    if args.chromosomes:
        ma.reorderChromosomes(args.chromosomes)

    # mask all zero value bins
    row_sum = np.asarray(ma.matrix.sum(axis=1)).flatten()
    log.info("Removing {} zero value bins".format(sum(row_sum == 0)))
    ma.maskBins(np.flatnonzero(row_sum == 0))
    matrix_shape = ma.matrix.shape

    if 'plotName' in args:
        plot_total_contact_dist(ma, args)
        sys.stderr.write("Saving diagnostic plot {}\n".format(args.plotName))
        exit()

    log.info("matrix contains {} data points. Sparsity {:.3f}.".format(
        len(ma.matrix.data),
        float(len(ma.matrix.data)) / (ma.matrix.shape[0]**2)))

    if args.skipDiagonal:
        ma.diagflat(value=0)

    outlier_regions = filter_by_zscore(ma, args.filterThreshold[0], args.filterThreshold[1], perchr=args.perchr)
    # compute and print some statistics
    pct_outlier = 100 * float(len(outlier_regions)) / ma.matrix.shape[0]
    ma.printchrtoremove(outlier_regions, label="Bins that are MAD outliers ({:.2f}%) "
                                               "out of".format(pct_outlier, ma.matrix.shape[0]),
                        restore_masked_bins=False)

    assert matrix_shape == ma.matrix.shape
    # mask filtered regions
    ma.maskBins(outlier_regions)
    total_filtered_out = set(outlier_regions)

    if args.sequencedCountCutoff and 0 < args.sequencedCountCutoff < 1:
        chrom, _, _, coverage = zip(*ma.cut_intervals)

        assert type(coverage[0]) == np.float64

        failed_bins = np.flatnonzero(
            np.array(coverage) < args.sequencedCountCutoff)

        ma.printchrtoremove(failed_bins, label="Bins with low coverage", restore_masked_bins=False)
        ma.maskBins(failed_bins)
        total_filtered_out = set(failed_bins)
        """
        ma.matrix, to_remove = fill_gaps(ma, failed_bins)
        sys.stderr.write("From {} failed bins, {} could "
                         "not be filled\n".format(len(failed_bins),
                                                  len(to_remove)))
        ma.maskBins(to_remove)
        """

    if args.transCutoff and 0 < args.transCutoff < 100:
        cutoff = float(args.transCutoff) / 100
        # a usual cutoff is 0.05
        ma.truncTrans(high=cutoff)

    pre_row_sum = np.asarray(ma.matrix.sum(axis=1)).flatten()
    correction_factors = []
    if args.perchr:
        corrected_matrix = lil_matrix(ma.matrix.shape)
        # normalize each chromosome independently
        for chrname in ma.interval_trees.keys():
            chr_range = ma.getChrBinRange(chrname)
            chr_submatrix = ma.matrix[chr_range[0]:chr_range[1], chr_range[0]:chr_range[1]]
            _matrix, _corr_factors = iterative_correction(chr_submatrix, args)
            corrected_matrix[chr_range[0]:chr_range[1], chr_range[0]:chr_range[1]] = _matrix
            correction_factors.append(_corr_factors)
        correction_factors = np.concatenate(correction_factors)

    else:
        corrected_matrix, correction_factors = iterative_correction(ma.matrix, args)

    ma.setMatrixValues(corrected_matrix)
    ma.setCorrectionFactors(correction_factors)
    if args.inflationCutoff and args.inflationCutoff > 0:
        after_row_sum = np.asarray(corrected_matrix.sum(axis=1)).flatten()
        # identify rows that were expanded more than args.inflationCutoff times
        to_remove = np.flatnonzero(after_row_sum / pre_row_sum >= args.inflationCutoff)
        ma.printchrtoremove(to_remove,
                            label="inflated >={} "
                            "regions".format(args.inflationCutoff), restore_masked_bins=False)
        total_filtered_out = total_filtered_out.union(to_remove)

        ma.maskBins(to_remove)

    ma.printchrtoremove(sorted(list(total_filtered_out)),
                        label="Total regions to be removed", restore_masked_bins=False)

    ma.save(args.outFileName)



from scipy.sparse import coo_matrix, dia_matrix, triu
import numpy as np
import time
import logging


def reduce_matrix(matrix, bins_to_merge, use_triu=True, diagonal=False):
    """
    This function sums the rows and columns corresponding
    to the bins_to_merge, returning a new sparse
    matrix of size len(bins_to_merge).

    The function uses sparse matrices tricks to work very fast.
    The idea is that all the cells in the new reduced matrix
    whose sum needs to be computed are expressed as a vector
    of bins.  Using the numpy bincount function the sum of each bin
    is obtained. Then this sum vector is converted to a
    sparse matrix.

    Parameters
    ----------

    bins_to_merge : A list of lists. The values of the lists should
        correspond to the indices of the matrix.

    use_triu : use only the upper triangle of the matrix. This is to avoid double
               counting values close to the diagonal.
               E.g. the matrix [[1, 2],
                                 2, 1]]

               when merging the two bins, should be reduced to [4] and not [6], because
               the number '2' should be counted only once.
               When this option is used, the result is converted to a symmetric matrix

    diagonal : If set to true, then the main diagonal is preserved (not deleted). Only works
               when use_triu is set

    Returns
    -------

    A sparse matrix.

    >>> from scipy.sparse import csr_matrix
    >>> A = csr_matrix(np.array([[1,0],[0,1]]))
    >>> reduce_matrix(A, [(0,1)], diagonal=True).todense()
    matrix([[2]])
    >>> A = csr_matrix(np.array([[5,5,2,2,0],[0,5,2,2,1],
    ... [0,0,1,1,0], [0,0,0,1,0], [0,0,0,0,0]]))
    >>> A.todense()
    matrix([[5, 5, 2, 2, 0],
            [0, 5, 2, 2, 1],
            [0, 0, 1, 1, 0],
            [0, 0, 0, 1, 0],
            [0, 0, 0, 0, 0]])
    >>> ll = [(0,1), (2,3), (4,)]
    >>> reduce_matrix(A, ll, diagonal=True, use_triu=False).todense()
    matrix([[15,  8,  1],
            [ 0,  3,  0],
            [ 0,  0,  0]])

    >>> ll = [(0,1,2), (3,4)]
    >>> reduce_matrix(A, ll, diagonal=True, use_triu=False).todense()
    matrix([[20,  6],
            [ 0,  1]])

    >>> ll = [(0,1), (2,3), (4,)]
    >>> reduce_matrix(A, ll, diagonal=True, use_triu=True).todense()
    matrix([[15,  8,  1],
            [ 8,  3,  0],
            [ 1,  0,  0]])

    Test symmetric matrix
    >>> A = csr_matrix(np.array([[2,2,1,1,1,1],[2,2,1,1,1,1],
    ... [1,1,1,1,1,1], [1,1,1,1,1,1], [1,1,1,1,1,1], [1,1,1,1,1,1]]))
    >>> A.todense()
    matrix([[2, 2, 1, 1, 1, 1],
            [2, 2, 1, 1, 1, 1],
            [1, 1, 1, 1, 1, 1],
            [1, 1, 1, 1, 1, 1],
            [1, 1, 1, 1, 1, 1],
            [1, 1, 1, 1, 1, 1]])
    >>> ll = [(0,1), (2,3), (4,5)]
    >>> reduce_matrix(A, ll, diagonal=True, use_triu=False).todense()
    matrix([[8, 4, 4],
            [4, 4, 4],
            [4, 4, 4]])

    >>> reduce_matrix(A, ll, diagonal=True, use_triu=True).todense()
    matrix([[6, 4, 4],
            [4, 3, 4],
            [4, 4, 3]])

    >>> ll = [(0,1,2), (3,4,5)]
    >>> reduce_matrix(A, ll, diagonal=True, use_triu=False).todense()
    matrix([[13,  9],
            [ 9,  9]])

    Test symmetric matrix non consecutive
    >>> A = csr_matrix(np.array([[5,1,5,1,0],[0,1,2,1,1],
    ... [0,0,5,1,0], [0,0,0,1,0], [0,0,0,0,0]]))
    >>> dia = dia_matrix(([A.diagonal()], [0]), shape=A.shape)
    >>> A= csr_matrix(A + A.T - dia)
    >>> A.todense()
    matrix([[5, 1, 5, 1, 0],
            [1, 1, 2, 1, 1],
            [5, 2, 5, 1, 0],
            [1, 1, 1, 1, 0],
            [0, 1, 0, 0, 0]])
    >>> ll = [(0,2), (1,3), (4,)]
    >>> reduce_matrix(A, ll, diagonal=True, use_triu=False).todense()
    matrix([[20,  5,  0],
            [ 5,  4,  1],
            [ 0,  1,  0]])

    Test removal of row/columns when the scaffold list does
    not contains all indices
    >>> ll = [(1,2), (3,)]
    >>> reduce_matrix(A, ll, diagonal=True, use_triu=False).todense()
    matrix([[10,  2],
            [ 2,  1]])

    Test with float and nan
    >>> A = csr_matrix(np.array([[0.1,0.1,0.2,0.2,np.nan],
    ... [0,0.1,0.2,0.2,1.1],
    ... [0,0,0.2,0.2,0], [0,0,0,0.1,0], [0,0,0,0,0]]))
    >>> dia = dia_matrix(([A.diagonal()], [0]), shape=A.shape)
    >>> A= csr_matrix(A + A.T - dia)
    >>> A.todense()
    matrix([[ 0.1,  0.1,  0.2,  0.2,  nan],
            [ 0.1,  0.1,  0.2,  0.2,  1.1],
            [ 0.2,  0.2,  0.2,  0.2,  0. ],
            [ 0.2,  0.2,  0.2,  0.1,  0. ],
            [ nan,  1.1,  0. ,  0. ,  0. ]])

    >>> ll = [(0,1), (2,3), (4,)]
    >>> reduce_matrix(A, ll, diagonal=True, use_triu=False).todense()
    matrix([[ 0.4,  0.8,  nan],
            [ 0.8,  0.7,  0. ],
            [ nan,  0. ,  0. ]])
    """

    if use_triu:
        ma = triu(matrix, k=0, format='coo')
    else:
        ma = matrix.tocoo()
    M = len(bins_to_merge)
    start_time = time.time()

    if M == ma.shape[0]:
        return matrix

    # check for nans and make a warning in case they are
    # present
    num_nan = len(np.flatnonzero(np.isnan(np.array(ma.data))))
    if num_nan > 0:
        logging.warn("*Warning*\nmatrix contains {} NaN values.".format(num_nan))

    # each original col and row index is converted
    # to a new index based on the bins_to_merge.
    # For example, if rows 1 and 10 are to be merged
    # then all rows whose value is 1 or 10 are given
    # as new value the index in the bins_to_merge list.

    map_ = np.zeros(ma.shape[0], dtype=int) - 1     # -1 such that all cases not replaced by the next loop
    # can be identified later. Those cases that remain as -1
    # are for the rows/cols not appearing in the bins_to_merge
    for k, v in enumerate(bins_to_merge):
        for x in v:
            map_[x] = k

    new_row = np.take(map_, ma.row)
    new_col = np.take(map_, ma.col)

    # remove rows and cols with -1. Those
    # correspond to regions that are not part of
    # the bins_to_merge to be merged
    keep = (new_row > -1) & (new_col > -1)
    ma.data = ma.data[keep]
    new_row = new_row[keep]
    new_col = new_col[keep]

    # The following line converts each combination
    # of row and col into a list of bins. For example,
    # for each case in which row = 1 and col =3 or
    # row = 3 and col = 1 a unique bin id is given
    # The trick I use to get the unique bin ids is
    # to order and convert the pair (row,col) into
    # a complex number  ( for the pair 3,1 the
    # complex es 1,3j.
    uniq, bin_array = np.unique(new_row + 1j * new_col, return_inverse=True)

    elapsed_time = time.time() - start_time
    start_time = time.time()
    logging.debug("time complex unique: {:.5f}".format(elapsed_time))
    # sum the bins. For each bin, all values associated
    # to the bin members are added. Those values are in
    # the ma.data array.
    sum_array = np.bincount(bin_array, weights=ma.data)

    # To reconstruct the new matrix the row and col
    # positions need to be obtained. They correspond
    # to the positions of the unique values in the bin_array.
    uniq, ind = np.unique(bin_array, return_index=True)
    new_row_ = new_row[ind]
    new_col_ = new_col[ind]

    result = coo_matrix((sum_array, (new_row_, new_col_)),
                        shape=(M, M), dtype=ma.dtype)
    elapsed_time = time.time() - start_time

    logging.debug("time create sparse matrix: {:.5f}".format(elapsed_time))

    if use_triu:
        diagmatrix = dia_matrix(([result.diagonal()], [0]),
                                shape=(M, M), dtype=ma.dtype)

        # to make set the main diagonal to zero I
        # multiply the diagonal by 2
        if diagonal is False:
            diagmatrix *= 2
        # make result symmetric
        matrix = result + result.T - diagmatrix
    else:
        matrix = result
    matrix.eliminate_zeros()

    return matrix



from __future__ import division
import sys
import argparse
from hicexplorer import HiCMatrix as hm
from hicexplorer._version import __version__
import numpy as np


def parse_arguments(args=None):
    """
    get command line arguments
    """
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description='Converts between different matrix file formats')

    # define the arguments
    parser.add_argument('--inFile', '-in',
                        help='input file(s). Could be one or many files. '
                        'Multiple input files are allowed for hicexplorer or lieberman format. '
                        ' In case of multiple input files, they will be combined. ',
                        nargs='+',
                        required=True)

    parser.add_argument('--inputFormat',
                        help='file format for input file. \n'
                             '(options : hicexplorer, lieberman, npz (file format of previous hicexplorer versions),'
                             'dekker.',
                        default='hicexplorer')

    parser.add_argument('--chrNameList',
                        help='list of chromosome names (only if input format is lieberman), eg : 1 2 .',
                        nargs='+',
                        )

    parser.add_argument('--outFileName', '-o',
                        help='File name to save the exported matrix. In the case of "lieberman" '
                             'output format this should be the path of a folder where the information '
                             'per chromosome is stored.',
                        required=True)

    parser.add_argument('--chromosomeOrder',
                        help='Chromosomes and order in which the chromosomes should be saved. If not all chromosomes '
                             'are given, those chromosomes are left out. For example, --chromosomeOrder chrX will '
                             'export a matrix only containing chromosome X',
                        nargs='+')

    parser.add_argument('--bplimit', '-b',
                        help='When merging many matrices : maximum limit (in base pairs) after '
                             'which the matrix will be truncated. i.e. TADs bigger than this '
                             'size will not be shown. For Matrices with very high resolution, '
                             'truncating the matrix after a limit helps in saving memory '
                             'during processing, without much loss of data. You can use '
                             'bplimit of 2 x size of biggest expected TAD. ',
                        type=int,
                        metavar='INT bp',
                        default=None)

    parser.add_argument('--outputFormat',
                        help='Output format. The possibilities are "dekker",  "ren", "hicexplorer, '
                             'npz (former hicexplorer format) and "GInteractoins". '
                             'The dekker format outputs the whole matrix where the '
                             'first column and first row are the bin widths and labels. '
                             'The "ren" format is a list of tuples of the form '
                             'chrom, bin_star, bin_end, values. '
                             'The lieberman format writes separate files for each chromosome,'
                             'with three columns : contact start, contact end, and raw observed score. '
                             'This corresponds to the RawObserved files from lieberman group. The '
                             'hicexplorer format stores the data using a hdf5 format. Optionally, '
                             'the numpy npz format can be used for small datasets (< 4GB).'
                             'The GInteractions format is in the form : Bin1, Bin2 , Interaction,'
                             'where Bin1 and Bin2 are intervals (chr,start,end), seperated by tab.',
                        default='dekker',
                        choices=['dekker', 'ren', 'lieberman', 'hicexplorer', 'npz', 'GInteractions'])

    parser.add_argument('--clearMaskedBins',
                        help='if set, masked bins are removed from the matrix. Masked bins '
                             'are those that do not have any values, mainly because they are'
                             'repetitive regions of the genome',
                        action='store_true')
    parser.add_argument('--version', action='version',
                        version='%(prog)s {}'.format(__version__))

    return parser


def combine_matrices(matrix_list, bplimit=None):
    """
    Combines individual matrices, stored per chromosome into
    one matrix

    :param matrix_list: name of the matrices that will be combined into one.
    :param bplimit: To reduce the final file size, counts over the given distance can be removed
    :return: sparse matrix, bin intervals, nan bins, corrrections factors, distance counts
    """

    # Create empty row, col, value for the matrix
    from scipy.sparse import coo_matrix, triu
    new_cut_intervals = []
    row = np.array([]).astype("int")
    col = np.array([]).astype("int")
    values = np.array([])
    new_nan_bins = np.array([]).astype('int')
    new_correction_factors = np.array([])
    new_distance_counts = np.array([])

    # for each chr, append the row, col, value to the first one. Extend the dim
    size = 0
    for i in range(0, len(matrix_list)):
        hic = hm.hiCMatrix(matrix_list[i])

        # trim matrix if bplimit given
        if bplimit is not None:
            limit = int(bplimit / hic.getBinSize())
            matrix = (triu(hic.matrix, k=-limit) - triu(hic.matrix, k=limit)).tocoo()
        else:
            matrix = hic.matrix.tocoo()

        # add data
        row = np.concatenate([row, matrix.row + size])
        col = np.concatenate([col, matrix.col + size])
        values = np.concatenate([values, matrix.data])
        new_nan_bins = np.concatenate([new_nan_bins, hic.nan_bins + size])
        new_cut_intervals.extend(hic.cut_intervals)
        size += matrix.shape[0]

        # also add correction_factors
        if hic.correction_factors is not None:
            new_correction_factors = np.append(new_correction_factors, hic.correction_factors)
        else:
            # add an array with NaNs
            arr = np.empty(matrix.shape[0])
            arr[:] = np.NAN
            new_correction_factors = np.concatenate([new_correction_factors, arr])
        if hic.distance_counts is not None:
            new_distance_counts = np.concatenate([new_distance_counts, hic.distance_counts])

    final_mat = coo_matrix((values, (row, col)), shape=(size, size)).tocsr()

    assert len(new_cut_intervals) == final_mat.shape[0], \
        "Corrupted matrix file. Matrix size and " \
        "matrix bin definitions do not correspond"

    if len(new_distance_counts) == 0:
        new_distance_counts = None
    if len(new_correction_factors) == 0:
        new_correction_factors = None

    return final_mat, new_cut_intervals, new_nan_bins, new_correction_factors, new_distance_counts


def main():
    args = parse_arguments().parse_args()

    # create hiC matrix with given input format
    # additional file needed for lieberman format
    if args.inputFormat == 'lieberman':
        if args.chrNameList is None:
            exit("Error: --chrNameList is required when the input format is lieberman. ")
        else:
            hic_ma = hm.hiCMatrix(matrixFile=args.inFile, file_format='lieberman', chrnameList=args.chrNameList)

    elif args.inputFormat == 'npz' and len(args.inFile) > 1:  # assume hicexplorer_multi format
        if args.bplimit:
            sys.stderr.write("\nCutting maximum matrix depth to {} for saving\n".format(args.bplimit))

        matrix, cut_intervals, nan_bins, corrections_factors, distance_counts = \
            combine_matrices(args.inFile, bplimit=args.bplimit)
        hic_ma = hm.hiCMatrix()
        hic_ma.setMatrix(matrix, cut_intervals=cut_intervals)

        if len(nan_bins):
            hic_ma.nan_bins = nan_bins
        if corrections_factors is not None:
            hic_ma.correction_factors = corrections_factors
        if distance_counts is not None:
            hic_ma.distance_counts = distance_counts

    else:
        hic_ma = hm.hiCMatrix(matrixFile=args.inFile[0], file_format=args.inputFormat)
        if args.bplimit:
            from scipy.sparse import triu
            sys.stderr.write("\nCutting maximum matrix depth to {} for saving\n".format(args.bplimit))

            limit = int(args.bplimit / hic_ma.getBinSize())
            hic_ma.matrix = (triu(hic_ma.matrix, k=-limit) - triu(hic_ma.matrix, k=limit)).tocsr()
            hic_ma.matrix.eliminate_zeros()

    if args.chromosomeOrder:
        hic_ma.keepOnlyTheseChr(args.chromosomeOrder)

    if args.clearMaskedBins:
        hic_ma.maskBins(hic_ma.nan_bins)

    sys.stderr.write('saving...\n')

    if args.outputFormat == 'dekker':
        hic_ma.save_dekker(args.outFileName)
    elif args.outputFormat == 'ren':
        hic_ma.save_bing_ren(args.outFileName)
    elif args.outputFormat == 'lieberman':
        hic_ma.save_lieberman(args.outFileName)
    elif args.outputFormat == 'npz':
        hic_ma.save_npz(args.outFileName)
    elif args.outputFormat == 'GInteractions':
        hic_ma.save_GInteractions(args.outFileName)
    else:
        hic_ma.save(args.outFileName)



import sys
import hicexplorer.HiCMatrix as HiCMatrix
from hicexplorer.utilities import writableFile
from hicexplorer._version import __version__
import numpy as np

import argparse
import matplotlib
matplotlib.use('Agg')

from matplotlib.colors import LogNorm
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import matplotlib.gridspec as gridspec


def parse_arguments(args=None):
    parser = argparse.ArgumentParser(description='Creates a Heatmap of a HiC matrix')

    # define the arguments
    parser.add_argument('--matrix', '-m',
                        help='Path of the Hi-C matrix to plot',
                        required=True)

    parser.add_argument('--title', '-t',
                        help='Plot title')

    parser.add_argument('--scoreName', '-s',
                        help='Score name')

    parser.add_argument('--outFileName', '-out',
                        help='File name to save the image. ',
                        type=writableFile,
                        required=True)

    parser.add_argument('--perChromosome',
                        help='Instead of plotting the whole matrix, '
                        'each chromosome is plotted next to the other. '
                        'This parameter is not compatible with --region',
                        action='store_true')

    parser.add_argument('--clearMaskedBins',
                        help='if set, masked bins are removed from the matrix',
                        action='store_true')

    parser.add_argument('--whatToShow',
                        help='Options are: "heatmap", "3D", and "both". '
                        'Default is heatmap',
                        default="heatmap",
                        choices=["heatmap", "3D", "both"])

    parser.add_argument('--chromosomeOrder',
                        help='Chromosomes and order in which the '
                        'chromosomes should be plotted. This option '
                        'overrides --region and --region2 ',
                        nargs='+')

    parser.add_argument('--region',
                        help='Plot only this region. The format is '
                        'chr:start-end The plotted region contains '
                        'the main diagonal and is symmetric unless '
                        ' --region2 is given'
                        )

    parser.add_argument('--region2',
                        help='If given, then only the region defined by '
                        '--region and --region2 is given. The format '
                        'is the same as --region1'
                        )

    parser.add_argument('--log1p',
                        help='Plot the log1p of the matrix values.',
                        action='store_true')

    parser.add_argument('--log',
                        help='Plot the *MINUS* log of the matrix values.',
                        action='store_true')

    parser.add_argument('--colorMap',
                        help='Color map to use for the heatmap. Available '
                        'values can be seen here: '
                        'http://matplotlib.org/examples/color/colormaps_reference.html',
                        default='RdYlBu_r')

    parser.add_argument('--vMin',
                        help='vMin',
                        type=float,
                        default=None)

    parser.add_argument('--vMax',
                        help='vMax',
                        type=float,
                        default=None)

    parser.add_argument('--zMax',
                        help='zMax for 3D plot',
                        type=float,
                        default=None)

    parser.add_argument('--dpi',
                        help='Resolution for the image in case the'
                             'ouput is a raster graphics image (e.g png, jpg)',
                        type=int,
                        default=72)

    parser.add_argument('--version', action='version',
                        version='%(prog)s {}'.format(__version__))

    return parser


def plotHeatmap3D(ma, fig, position, args, cmap):
    axHeat1 = fig.add_axes(position, projection='3d')
    axHeat1.view_init(50, -45)
    axHeat1.margins(0)

    X, Y = np.meshgrid(range(ma.shape[0]), range(ma.shape[0]))
    ma = np.ma.array(ma, mask=np.isnan(ma))
    Z = ma.copy()

    Z[np.where(np.tril(Z) == 0)] = np.nan

    axHeat1.plot_surface(X, Y, Z, rstride=1, cstride=1, linewidth=0, cmap=cmap,
                         vmax=args.vMax, vmin=args.vMin)
    if args.zMax:
        axHeat1.set_zlim(0, args.zMax)


def change_chrom_names(chrom):
    """
    Changes UCSC chromosome names to ensembl chromosome names
    and vice versa.
    """
    # TODO: mapping from chromosome names like mithocondria is missing
    if chrom.startswith('chr'):
        # remove the chr part from chromosome name
        chrom = chrom[3:]
    else:
        # prefix with 'chr' the chromosome name
        chrom = 'chr' + chrom

    return chrom


def plotHeatmap(ma, chrBinBoundaries, fig, position, args, figWidth, cmap):

    axHeat2 = fig.add_axes(position)
    if args.title:
        axHeat2.set_title(args.title)
    norm = None
    if args.log1p:
        ma += 1
        norm = LogNorm()

    img3 = axHeat2.imshow(ma,
                          interpolation='nearest',
                          #                          interpolation='spline16',
                          vmax=args.vMax, vmin=args.vMin, cmap=cmap,
                          norm=norm
                          )

    img3.set_rasterized(True)
    from mpl_toolkits.axes_grid1 import make_axes_locatable
    divider = make_axes_locatable(axHeat2)
    cax = divider.append_axes("right", size="2.5%", pad=0.09)
    cbar = fig.colorbar(img3, cax=cax)
    cbar.solids.set_edgecolor("face")  # to avoid white lines in the color bar in pdf plots
    if args.scoreName:
        cbar.ax.set_ylabel(args.scoreName, rotation=270, size=8)

    if args.region:
        """
        ticks = axHeat2.get_xticks()
        labels = ["{:.3f} Mbp".format((x / 1e6))
                  for x in ticks]
        axHeat2.get_xaxis().set_tick_params(
            which='both',
            bottom='on')
        axHeat2.set_xticklabels(labels, size='small')
        """

        axHeat2.set_xticks([0, ma.shape[0]])
        axHeat2.set_xticklabels([args.region[1], args.region[2]], size=4, rotation=90)
        axHeat2.set_axis_off()
    else:
        ticks = [int(pos[0] + (pos[1] - pos[0]) / 2) for pos in chrBinBoundaries.values()]
        labels = chrBinBoundaries.keys()
        axHeat2.set_xticks(ticks)
        if len(labels) > 20:
            axHeat2.set_xticklabels(labels, size=4, rotation=90)
        else:
            axHeat2.set_xticklabels(labels, size=8)

    axHeat2.get_yaxis().set_visible(False)


def plotHeatmap_region(ma, chrBinBoundaries, fig, position, args, cmap, xlabel=None,
                       ylabel=None, start_pos=None, start_pos2=None):

    axHeat2 = fig.add_axes(position)
    if args.title:
        axHeat2.set_title(args.title)
    norm = None
    if args.log1p:
        ma += 1
        norm = LogNorm()

    if start_pos is None:
        start_pos = np.arange(ma.shape[0])
    if start_pos2 is None:
        start_pos2 = start_pos

    xmesh, ymesh = np.meshgrid(start_pos, start_pos2)
    img3 = axHeat2.pcolormesh(xmesh.T, ymesh.T, ma, vmin=args.vMin, vmax=args.vMax, cmap=cmap, norm=norm)

    img3.set_rasterized(True)

    if args.region:

        # relabel xticks
        def relabel_ticks(ticks):
            if ticks[-1] - ticks[0] > 100000:
                labels = ["{:.2f} ".format((x / 1e6))
                          for x in ticks]
                labels[-1] += "Mbp"
            else:
                labels = ["{:,} ".format((x))
                          for x in ticks]
                labels[-1] += "bp"
            return labels
        xtick_lables = relabel_ticks(axHeat2.get_xticks())
        axHeat2.get_xaxis().set_tick_params(which='both', bottom='on', direction='out')
        axHeat2.set_xticklabels(xtick_lables, size='small', rotation=45)

        ytick_lables = relabel_ticks(axHeat2.get_yticks())
        axHeat2.get_yaxis().set_tick_params(which='both', bottom='on', direction='out')
        axHeat2.set_yticklabels(ytick_lables, size='small')

        """
        axHeat2.set_xticks([0, ma.shape[0]])
        axHeat2.set_xticklabels([args.region[1], args.region[2]], size=4, rotation=90)
        axHeat2.set_axis_off()
        """
    else:
        ticks = [int(pos[0] + (pos[1] - pos[0]) / 2) for pos in chrBinBoundaries.values()]
        labels = chrBinBoundaries.keys()
        axHeat2.set_xticks(ticks)
        if len(labels) > 20:
            axHeat2.set_xticklabels(labels, size=4, rotation=90)
        else:
            axHeat2.set_xticklabels(labels, size=8)

    from mpl_toolkits.axes_grid1 import make_axes_locatable
    divider = make_axes_locatable(axHeat2)
    cax = divider.append_axes("right", size="2.5%", pad=0.09)
    if args.log1p:
        from matplotlib.ticker import LogFormatter
        formatter = LogFormatter(10, labelOnlyBase=False)
        # get a useful log scale
        # that looks like [1, 2, 5, 10, 20, 50, 100, ... etc]
        aa = np.array([1, 2, 5])
        tick_values = np.concatenate([aa * 10**x for x in range(10)])
        cbar = fig.colorbar(img3, ticks=tick_values, format=formatter, cax=cax)
    else:
        cbar = fig.colorbar(img3, cax=cax)

    cbar.solids.set_edgecolor("face")  # to avoid white lines in the color bar in pdf plots
    if args.scoreName:
        cbar.ax.set_ylabel(args.scoreName, rotation=270, size=8)
    axHeat2.set_ylim(start_pos2[0], start_pos2[-1])
    axHeat2.set_xlim(start_pos[0], start_pos[-1])
    axHeat2.spines['top'].set_visible(False)
    axHeat2.spines['right'].set_visible(False)

    if ylabel is not None:
        axHeat2.set_ylabel(ylabel)

    if xlabel is not None:
        axHeat2.set_xlabel(xlabel)


def translate_region(region_string):
    """
    Takes an string and returns a list
    of chrom, start, end.
    If the region string only contains
    the chrom, then start and end
    are set to a 0 and 1e15
    """

    region_string = region_string.translate(None, ",;!").replace("-", ":")
    fields = region_string.split(":")
    chrom = fields[0]
    try:
        region_start = int(fields[1])
    except IndexError:
        region_start = 0
    try:
        region_end = int(fields[2])
    except IndexError:
        region_end = 1e15  # vert large number

    return chrom, region_start, region_end


def plotPerChr(hic_matrix, cmap, args):
    """
    plots each chromosome individually, one after the other
    in one row. scale bar is added at the end
    """
    from math import ceil
    chromosomes = hic_matrix.getChrNames()
    chrom_per_row = 5
    num_rows = int(ceil(float(len(chromosomes)) / chrom_per_row))
    num_cols = min(chrom_per_row, len(chromosomes))
    width_ratios = [1] * num_cols + [0.05]
    grids = gridspec.GridSpec(num_rows, num_cols + 1,
                              width_ratios=width_ratios,
                              height_ratios=[1] * num_rows)

    fig_height = 6 * num_rows
    fig_width = sum((np.array(width_ratios) + 0.05) * 6)

    fig = plt.figure(figsize=(fig_width, fig_height), dpi=args.dpi)

    norm = None
    if args.log1p:
        norm = LogNorm()
    chrom, start, end, _ = zip(*hic_matrix.cut_intervals)

    for idx, chrname in enumerate(chromosomes):
        row = idx / chrom_per_row
        col = idx % chrom_per_row

        axis = plt.subplot(grids[row, col])
        axis.set_title(chrname)
        chrom_range = hic_matrix.getChrBinRange(chrname)
        mat = hic_matrix.matrix[chrom_range[0]:chrom_range[1],
                                chrom_range[0]:chrom_range[1]].todense().astype(float)

        if args.log1p:
            mat += 1
        img = axis.imshow(mat, aspect='auto',
                          interpolation='spline16',
                          vmax=args.vMax, vmin=args.vMin, cmap=cmap,
                          norm=norm,
                          extent=[start[chrom_range[0]], end[chrom_range[1] - 1],
                                  end[chrom_range[1] - 1], start[chrom_range[0]]])

        img.set_rasterized(True)

        xticks = axis.get_xticks()
        if xticks[1] < 1e6:
            xlabels = ["{:.0f}".format(int(x) / 1e3)
                       for x in xticks]
            xlabels[-2] = "{} Kb".format(xlabels[-2])
        else:
            xlabels = ["{:.0f}".format(int(x) / 1e6)
                       for x in xticks]
            xlabels[-2] = "{} Mb".format(xlabels[-2])

        axis.set_xticklabels(xlabels, size='small')
        # yticks = axis.get_yticks()

        # ylabels = ["{:.0f}".format(int(x) / 1e6)
        #          for x in yticks]

        axis.get_xaxis().set_tick_params(
            which='both',
            bottom='on',
            direction='out')

        axis.set_yticklabels(xlabels, size='small')

    cbar3 = plt.subplot(grids[-1])
    cbar = fig.colorbar(img, cax=cbar3)
    cbar.solids.set_edgecolor("face")  # to avoid white lines in
    # the color bar in pdf plots
    cbar.ax.set_ylabel(args.scoreName, rotation=270, labelpad=20)


def main(args=None):
    args = parse_arguments().parse_args(args)

    ma = HiCMatrix.hiCMatrix(args.matrix)
    if args.perChromosome and args.region:
        sys.stderr.write('ERROR, choose from the option '
                         '--perChromosome or --region, the two '
                         'options at the same time are not '
                         'compatible.')
    if args.chromosomeOrder:
        args.region = None
        args.region2 = None
        ma.reorderChromosomes(args.chromosomeOrder)

    print ma.interval_trees.keys()
    ma.restoreMaskedBins()

    if args.clearMaskedBins:
        ma.maskBins(ma.nan_bins)

    sys.stderr.write("min: {}, max: {}\n".format(ma.matrix.data.min(), ma.matrix.data.max()))
    if args.region:
        chrom, region_start, region_end = translate_region(args.region)
        if chrom not in ma.interval_trees.keys():
            chrom = change_chrom_names(chrom)
            if chrom not in ma.interval_trees.keys():
                exit("Chromosome name {} in --region not in matrix".format(change_chrom_names(chrom)))

        args.region = [chrom, region_start, region_end]
        idx1, start_pos1 = zip(*[(idx, x[1]) for idx, x in enumerate(ma.cut_intervals) if x[0] == chrom and
                                 x[1] >= region_start and x[2] < region_end])
        if args.region2:
            chrom2, region_start2, region_end2 = translate_region(args.region2)
            if chrom2 not in ma.interval_trees.keys():
                chrom2 = change_chrom_names(chrom2)
                if chrom2 not in ma.interval_trees.keys():
                    exit("Chromosome name {} in --region2 not in matrix".format(change_chrom_names(chrom2)))
            idx2, start_pos2 = zip(*[(idx, x[1]) for idx, x in enumerate(ma.cut_intervals) if x[0] == chrom2 and
                                     x[1] >= region_start2 and x[2] < region_end2])
        else:
            idx2 = idx1
            chrom2 = chrom
            start_pos2 = start_pos1
        # select only relevant part
        matrix = np.asarray(ma.matrix[idx1, :][:, idx2].todense().astype(float))

    else:
        # TODO make start_pos1
        start_pos1 = None
        start_pos2 = None
        chrom = None
        chrom2 = None
        matrix = np.asanyarray(ma.getMatrix().astype(float))

    cmap = cm.get_cmap(args.colorMap)
    sys.stderr.write("Nan values set to black\n")
    cmap.set_bad('black')

    if args.perChromosome:
        plotPerChr(ma, cmap, args)

    else:
        # set as nan regions the nan bins

        fig_height = 6
        height = 5.0 / fig_height
        if args.whatToShow == 'both':
            fig_width = 11
            width = 4.9 / fig_width
            left_margin = (1.0 - 2 * width) * 0.35
        else:
            fig_width = 7
            width = 4.4 / fig_width
            left_margin = (1.0 - width) * 0.35

        fig = plt.figure(figsize=(fig_width, fig_height), dpi=args.dpi)
        bottom = 0.7 / fig_height

        if args.log:
            mask = matrix == 0
            matrix[mask] = matrix[mask is False].min()
    #        matrix = -1 *  np.log(matrix)
            matrix = np.log(matrix)

        if args.whatToShow == '3D':
            position = [left_margin, bottom, width * 1.2, height * 1.2]
            plotHeatmap3D(matrix, fig, position, args, cmap)

        if args.whatToShow == 'both':
            position = [left_margin, bottom, width * 1.2, height * 1.2]
            plotHeatmap3D(matrix, fig, position, args, cmap)

            left_margin2 = 5.5 / fig_width + left_margin
            position = [left_margin2, bottom, width, height]
            plotHeatmap(matrix, ma.chrBinBoundaries, fig, position, args,
                        fig_width, cmap)

        if args.whatToShow == 'heatmap':
            position = [left_margin, bottom, width, height]
            if args.region:
                plotHeatmap_region(matrix, ma.chrBinBoundaries, fig, position,
                                   args, cmap, xlabel=chrom, ylabel=chrom2,
                                   start_pos=start_pos1, start_pos2=start_pos2)

            else:
                plotHeatmap(matrix, ma.chrBinBoundaries, fig, position,
                            args, fig_width, cmap)

    plt.savefig(args.outFileName, dpi=args.dpi)



from __future__ import division
import argparse
from hicexplorer import HiCMatrix as hm
from hicexplorer._version import __version__
import numpy as np


def parse_arguments(args=None):
    """
    get command line arguments
    """
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description='Converts between different matrix file formats')

    # define the arguments
    parser.add_argument('--treatment', '-t',
                        help='Treatment Hi-C matrix',
                        required=True)

    parser.add_argument('--control', '-c',
                        help='Control Hi-C matrix',
                        required=True)

    parser.add_argument('--outFileName', '-o',
                        help='File name to save the resulting matrix',
                        required=True)

    parser.add_argument('--version', action='version',
                        version='%(prog)s {}'.format(__version__))

    return parser


def main():
    args = parse_arguments().parse_args()

    hic_t = hm.hiCMatrix(matrixFile=args.treatment)
    hic_c = hm.hiCMatrix(matrixFile=args.control)

    # scale larger matrix down
    total_t = hic_t.matrix.sum()
    total_c = hic_c.matrix.sum()

    if total_c > total_t:
        scale_factor = [1, float(total_t) / total_c]
    else:
        scale_factor = [float(total_c) / total_t, 1]

    hic_t.matrix.data = hic_t.matrix.data * scale_factor[0]
    hic_c.matrix.data = hic_c.matrix.data * scale_factor[1]

    """
    Uses sparse matrix tricks to convert
    into a vector the matrix values such
    that zero values that appear in only
    one of the matrices is kept. But
    zeros in two matrices are removed

    Requires two sparse matrices as input
    """
    assert hic_t.matrix.shape == hic_c.matrix.shape, "Matrices have different shapes."

    assert (hic_t.matrix - hic_c.matrix).sum() != 0, "Matrices are identical."

    # create a new matrix that is the sum of the two
    # matrices to compare. The goal is to have
    # a matrix that contains all the positions
    # that are non-zero in both matrices
    _mat = hic_t.matrix + hic_c.matrix

    # add one to each element in the new matrix
    _mat.data += 1

    # get a vector of the values in hic_t from
    # _mat
    values_t = (_mat - hic_t.matrix).data - 1

    # get a vector of the values in hic_c from
    # _mat
    values_c = (_mat - hic_c.matrix).data - 1

    # compute log2ratio
    _mat.data = np.log2(values_t / values_c)

    hic_t.matrix = _mat
    hic_t.matrix.eliminate_zeros()
    hic_t.save(args.outFileName)



import sys
import argparse
import numpy as np

from hicexplorer import HiCMatrix as hm
from hicexplorer.reduceMatrix import reduce_matrix
from hicexplorer._version import __version__

debug = 0


def parse_arguments(args=None):

    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description='Merges bins from a Hi-C matrix. For example, '
        'using a matrix containing 5bk bins, a matrix '
        'of 50 kb bins can be derived. ')

    # define the arguments
    parser.add_argument('--matrix', '-m',
                        help='Matrix to reduce.',
                        metavar='.h5 fileformat',
                        required=True)

    parser.add_argument('--numBins', '-nb',
                        help='Number of bins to merge.',
                        metavar='int',
                        type=int,
                        required=True)

    parser.add_argument('--runningWindow',
                        help='set to merge for using a running '
                        'window of length --numBins',
                        action='store_true')

    parser.add_argument('--outFileName', '-o',
                        help='File name to save the resulting matrix. '
                        'The output is also a .h5 file. But don\'t add '
                        'the suffix',
                        required=True)

    parser.add_argument('--version', action='version',
                        version='%(prog)s {}'.format(__version__))

    return parser


def running_window_merge(hic_matrix):
    """Creates a 'running window' merge. If the original matrix
    has bins as 5000 resolution the output matrix also has 5000 resolution.
    However, each element of the matrix is the sum of the 4 neighboring
    regions. For the followin matrix, the sums are as follows: a+b+d+e,
    b+c+d+f etc.

       a | b | c
       ---------
       d | e | f
       ---------
       g | h | i


    >>> from scipy.sparse import csr_matrix
    >>> row, col = np.triu_indices(5)
    >>> cut_intervals = [('a', 0, 10, 0.5), ('a', 10, 20, 1),
    ... ('a', 20, 30, 1), ('a', 30, 40, 0.1), ('b', 40, 50, 1)]
    >>> hic = hm.hiCMatrix()
    >>> hic.nan_bins = []
    >>> matrix = np.array([
    ... [ 0, 10,  5, 3, 0],
    ... [ 0,  0, 15, 5, 1],
    ... [ 0,  0,  0, 7, 3],
    ... [ 0,  0,  0, 0, 1],
    ... [ 0,  0,  0, 0, 0]])

    make the matrix symmetric:
    >>> hic.matrix = csr_matrix(matrix + matrix.T)
    >>> hic.setMatrix(csr_matrix(matrix + matrix.T), cut_intervals)
    >>> merge_matrix = running_window_merge(hic)
    >>> merge_matrix.cut_intervals
    [('a', 5, 15, 0.75), ('a', 15, 25, 1.0), ('a', 25, 35, 0.55000000000000004)]
    >>> merge_matrix.matrix.todense()
    matrix([[20, 30, 28],
            [30, 30, 27],
            [28, 27, 14]])
    """

    from scipy.sparse import csr_matrix, vstack, hstack

    num_rows = num_cols = hic_matrix.matrix.shape[0]
    # add an extra last row the original matrix
    rows_mat = csr_matrix((num_rows, 1))
    merge_matrix = hstack([hic_matrix.matrix, rows_mat])
    # add an extra last column to the original matrix
    cols_mat = csr_matrix((1, num_cols + 1))
    merge_matrix = vstack([merge_matrix, cols_mat],
                          format='coo').astype(int)

    # merge matrix is equal to merge_matrix + down + right + down_right
    # where down is the same matrix but all rows are +1,
    # for right, all cols are +1 and for down_right, cols and rows are
    # +1
    down = merge_matrix.copy()
    down.row += 1

    right = merge_matrix.copy()
    right.col += 1

    down_right = down.copy()
    down_right.col += 1

    merge_matrix = (merge_matrix + down + right + down_right).tocsr()
    # new matrix is one row and one column bigger, which need to be
    # removed
    merge_matrix = merge_matrix[1:, :][:, 1:]

    # not the intervals need to be adjusted.
    # the original interval 'chr1', 0, 5000' needs
    # to be rewritten as 'chr1', 2500, 7500
    bin_size_half = int(hic_matrix.getBinSize() / 2)
    to_keep = []
    total_intervals = len(hic_matrix.cut_intervals)
    new_intervals = []
    for idx, interval in enumerate(hic_matrix.cut_intervals):
        chrom, start, end, coverage = interval
        if idx + 1 >= total_intervals:
            continue
        if chrom != hic_matrix.cut_intervals[idx + 1][0]:
            # the next interval is for a different
            # chromosome. That means, this region has
            # to be removed
            continue
        to_keep.append(idx)
        coverage = np.mean([coverage, hic_matrix.cut_intervals[idx + 1][-1]])
        new_intervals.append((chrom, start + bin_size_half,
                              end + bin_size_half, coverage))

    hic_matrix.matrix = merge_matrix[to_keep, :][:, to_keep]
    hic_matrix.cut_intervals = new_intervals
    hic_matrix.nan_bins = np.flatnonzero(hic_matrix.matrix.sum(0).A == 0)

    return hic_matrix


def running_window_merge_v2(hic_matrix, num_bins):
    """Creates a 'running window' merge without changing the
    original resolution of the matrix. The window size is
    defined by the num_bins that are merged. Num bins
    had to be an odd number such that equal amounts of left and
    right bins can be merged.

       a | b | c
       ---------
       d | e | f
       ---------
       g | h | i

    In this matrix, using a merge of num_bins 3,
    the merge is done as follows, a = a + b + d + e,
    e = a + b + c + d + e + f etc,

    >>> from scipy.sparse import csr_matrix, dia_matrix
    >>> row, col = np.triu_indices(5)
    >>> cut_intervals = [('a', 0, 10, 0.5), ('a', 10, 20, 1),
    ... ('a', 20, 30, 1), ('a', 30, 40, 0.1), ('b', 40, 50, 1)]
    >>> hic = hm.hiCMatrix()
    >>> hic.nan_bins = []
    >>> matrix = np.array([
    ... [ 1, 1 ],
    ... [ 1, 1 ]])

    make the matrix symmetric:
    >>> hic.matrix = csr_matrix(matrix)
    >>> hic.setMatrix(hic.matrix, cut_intervals)
    >>> merge_matrix = running_window_merge_v2(hic, 3)
    >>> merge_matrix.matrix.todense()
    matrix([[3, 3],
            [3, 3]])

    >>> matrix = np.array([
    ... [ 1, 1, 1, 1 ],
    ... [ 1, 1, 1, 1 ],
    ... [ 1, 1, 1, 1 ],
    ... [ 1, 1, 1, 1 ]])

    make the matrix symmetric:
    >>> hic.matrix = csr_matrix(matrix)
    >>> hic.setMatrix(hic.matrix, cut_intervals)
    >>> merge_matrix = running_window_merge_v2(hic, 3)
    >>> merge_matrix.matrix.todense()
    matrix([[3, 5, 6, 4],
            [5, 6, 8, 6],
            [6, 8, 6, 5],
            [4, 6, 5, 3]])
    """

    if num_bins == 1:
        return hic_matrix

    assert num_bins % 2 == 1, "num_bins has to be an odd number"
    half_num_bins = int((num_bins - 1) / 2)
    from scipy.sparse import coo_matrix, dia_matrix, triu
    M = hic_matrix.matrix.shape[0]
    ma = triu(hic_matrix.matrix, k=0, format='coo')
    row = ma.row
    col = ma.col
    data = ma.data
    # indices list:
    idx_list = []
    for i in range(num_bins):
        for j in range(num_bins):
            idx_list.append((j - half_num_bins, i - half_num_bins))

    new_row = row
    new_col = col
    new_data = data
    for idx_pair in idx_list:
        if idx_pair == (0, 0):
            continue
        new_row = np.concatenate([new_row, row + idx_pair[0]])
        new_col = np.concatenate([new_col, col + idx_pair[1]])
        new_data = np.concatenate([new_data, data])

#    import ipdb;ipdb.set_trace()
#    print np.vstack([new_row, new_col, new_data])

    # remove illegal matrix id
    # that are less than zero
    # or bigger than matrix size
    keep = ((new_row > -1) & (new_col > -1) &
            (new_row < M) & (new_col < M))
    new_data = new_data[keep]
    new_row = new_row[keep]
    new_col = new_col[keep]

    new_ma = coo_matrix((new_data, (new_row, new_col)), shape=(M, M))
    new_ma = triu(new_ma, k=0)
#   new_ma.data = new_ma.data / len(idx_list)
    dia = dia_matrix(([new_ma.diagonal()], [0]), shape=new_ma.shape)
    new_ma = new_ma + new_ma.T - dia

    hic_matrix.matrix = new_ma
    hic_matrix.nan_bins = np.flatnonzero(hic_matrix.matrix.sum(0).A == 0)

    return hic_matrix


def merge_bins(hic, num_bins):
    """
    Merge the bins using the specified number of bins. This
    functions takes care to make new intervals

    Parameters
    ----------

    hic : HiCMatrix object

    num_bins : number of consecutive bins to merge.

    Returns
    -------

    A sparse matrix.

    Set up a Hi-C test matrix
    >>> from scipy.sparse import csr_matrix
    >>> row, col = np.triu_indices(5)
    >>> cut_intervals = [('a', 0, 10, 0.5), ('a', 10, 20, 1),
    ... ('a', 20, 30, 1), ('a', 30, 40, 0.1), ('b', 40, 50, 1)]
    >>> hic = hm.hiCMatrix()
    >>> hic.nan_bins = []
    >>> matrix = np.array([
    ... [ 50, 10,  5,  3,   0],
    ... [  0, 60, 15,  5,   1],
    ... [  0,  0, 80,  7,   3],
    ... [  0,  0,  0, 90,   1],
    ... [  0,  0,  0,  0, 100]])

    make the matrix symmetric:
    >>> from scipy.sparse import dia_matrix

    >>> dia = dia_matrix(([matrix.diagonal()], [0]), shape=matrix.shape)
    >>> hic.matrix = csr_matrix(matrix + matrix.T - dia)
    >>> hic.setMatrix(hic.matrix, cut_intervals)

    run merge_matrix
    >>> merge_matrix = merge_bins(hic, 2)
    >>> merge_matrix.cut_intervals
    [('a', 0, 20, 0.75), ('a', 20, 40, 0.55000000000000004), ('b', 40, 50, 1.0)]
    >>> merge_matrix.matrix.todense()
    matrix([[120,  28,   1],
            [ 28, 177,   4],
            [  1,   4, 100]])
    """
    # get the bins to merge
    ref_name_list, start_list, end_list, coverage_list = zip(*hic.cut_intervals)
    new_bins = []
    bins_to_merge = []
    prev_ref = ref_name_list[0]

    # prepare new intervals
    idx_start = 0
    new_start = start_list[0]
    count = 0
    for idx, ref in enumerate(ref_name_list):
        if (count > 0 and count % num_bins == 0) or ref != prev_ref:
            if count < num_bins / 2:
                sys.stderr.write("{} has few bins ({}). Skipping it\n".format(prev_ref, count))
            else:
                coverage = np.mean(coverage_list[idx_start:idx])
                new_bins.append((ref_name_list[idx_start], new_start, end_list[idx - 1], coverage))
                bins_to_merge.append(range(idx_start, idx))
            idx_start = idx
            new_start = start_list[idx]
            count = 0

        prev_ref = ref
        count += 1
    coverage = np.mean(coverage_list[idx_start:])
    new_bins.append((ref, new_start, end_list[idx], coverage))
    bins_to_merge.append(range(idx_start, idx + 1))

    hic.matrix = reduce_matrix(hic.matrix, bins_to_merge, diagonal=True)
    hic.cut_intervals = new_bins
    hic.nan_bins = np.flatnonzero(hic.matrix.sum(0).A == 0)

    return hic


def main():

    args = parse_arguments().parse_args()
    hic = hm.hiCMatrix(args.matrix)
    if args.runningWindow:
        merged_matrix = running_window_merge_v2(hic, args.numBins)
    else:
        merged_matrix = merge_bins(hic, args.numBins)

    print 'saving matrix'
    # there is a pickle problem with large arrays
    # To increase the sparsity of the matrix and
    # overcome the problem
    # I transform al ones into zeros.
    """
    merged_matrix.matrix.data = merged_matrix.matrix.data - 1
    """
    merged_matrix.matrix.eliminate_zeros()
    if merged_matrix.correction_factors is not None:
        sys.stderr.write("*WARNING*: The corrections factors are not merged and are set to None\n")
        merged_matrix.correction_factors = None

    merged_matrix.save(args.outFileName)



import numpy as np
import time
import logging

logging.basicConfig()
log = logging.getLogger("iterative correction")
log.setLevel(logging.WARN)


def iterativeCorrection(matrix, v=None, M=50, tolerance=1e-5, verbose=False):
    """
    adapted from cytonised version in mirnylab
    original code from: ultracorrectSymmetricWithVector
    https://bitbucket.org/mirnylab/mirnylib/src/924bfdf5ed344df32743f4c03157b0ce49c675e6/mirnylib/numutils_new.pyx?at=default

    Main method for correcting DS and SS read data.
    Possibly excludes diagonal.
    By default does iterative correction, but can perform an M-time correction
    :param matrix: a scipy sparse matrix
    :param tolerance: Tolerance is the maximum allowed relative
                      deviation of the marginals.
    """
    if verbose:
        log.setLevel(logging.INFO)

    total_bias = np.ones(matrix.shape[0], 'float64')

    if np.isnan(matrix.sum()):
        log.warn("[iterative correction] the matrix contains nans, they will be replaced by zeros\n")
        matrix.data[np.isnan(matrix.data)] = 0

    matrix = matrix.astype(float)
    W = matrix.tocoo()

    if np.abs(matrix - matrix.T).mean() / (1. * np.abs(matrix.mean())) > 1e-10:
        raise ValueError("Please provide symmetric matrix!")

    start_time = time.time()
    log.info("starting iterative correction")
    for iternum in xrange(M):
        iternum += 1
        s = np.array(W.sum(axis=1)).flatten()
        mask = (s == 0)
        s = s / np.mean(s[~mask])

        total_bias *= s
        deviation = np.abs(s - 1).max()

        s = 1.0 / s

        # The following code  is an optimization of this
        # for i in range(N):
        #     for j in range(N):
        #         W[i,j] = W[i,j] / (s[i] * s[j])

        W.data *= np.take(s, W.row)
        W.data *= np.take(s, W.col)
        if np.any(W.data > 1e100):
            log.error("*Error* matrix correction is producing extremely large values. "
                      "This is often caused by bins of low counts. Use a more stringent "
                      "filtering of bins.")
            exit(1)
        if verbose:
            if iternum % 5 == 0:
                end_time = time.time()
                estimated = (float(M - iternum) * (end_time - start_time)) / iternum
                m, sec = divmod(estimated, 60)
                h, m = divmod(m, 60)
                log.info("pass {} Estimated time {:.0f}:{:.0f}:{:.0f}".format(iternum, h, m, sec))
                log.info("max delta - 1 = {} ".format(deviation))

        if deviation < tolerance:
            log.info("[iterative correction] {} iterations used\n".format(iternum + 1))
            break

    # scale the total bias such that the sum is 1.0
    corr = total_bias[total_bias != 0].mean()
    total_bias /= corr
    W.data = W.data * corr * corr
    if np.any(W.data > 1e10):
        log.error("*Error* matrix correction produced extremely large values. "
                  "This is often caused by bins of low counts. Use a more stringent "
                  "filtering of bins.")
        exit(1)

    return W.tocsr(), total_bias



import argparse
from hicexplorer import HiCMatrix as hm
from hicexplorer._version import __version__


def parse_arguments(args=None):

    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                     description=('Adds Hi-C matrices of the same size. Format '
                                                  'has to be hdf5 or npz'))

    parser.add_argument('--matrices', '-m',
                        help='matrices to add. Must have the same shape.',
                        metavar='.h5 file format',
                        nargs='+',
                        required=True)

    parser.add_argument('--outFileName', '-o',
                        help='File name to save the resulting matrix. The output is '
                             'also a .h5 file. But don\'t add the suffix',
                        required=True)

    parser.add_argument('--version', action='version',
                        version='%(prog)s {}'.format(__version__))

    return parser


def main():

    args = parse_arguments().parse_args()
    hic = hm.hiCMatrix(args.matrices[0])
    summed_matrix = hic.matrix
    nan_bins = set(hic.nan_bins)
    for matrix in args.matrices[1:]:
        hic_to_append = hm.hiCMatrix(matrix)

        try:
            summed_matrix = summed_matrix + hic_to_append.matrix
            if len(hic_to_append.nan_bins):
                nan_bins = nan_bins.union(hic_to_append.nan_bins)
        except:
            print "\nMatrix {} seems to be corrupted or of different " \
                  "shape".format(matrix)
            exit(1)

    # save only the upper triangle of the
    # symmetric matrix
    hic.setMatrixValues(summed_matrix)
    hic.maskBins(sorted(nan_bins))
    hic.save(args.outFileName)
